import csv
import itertools
from datetime import datetime as Datetime
from decimal import Decimal

from werkzeug.exceptions import BadRequest

from chellow.utils import parse_channel_type, parse_mpan_core, to_utc, validate_hh_start


def create_parser(reader, mpan_map, messages):
    return HhParserCsvSimple(reader, mpan_map, messages)


class HhParserCsvSimple:
    def __init__(self, reader, mpan_map, messages):
        self.shredder = zip(itertools.count(1), csv.reader(reader))
        next(self.shredder)  # skip the title line
        self.values = None

    def get_field(self, index, name):
        if len(self.values) > index:
            return self.values[index].strip()
        else:
            raise BadRequest("Can't find field " + index + ", " + name + ".")

    def __iter__(self):
        return self

    def __next__(self):
        try:
            self.line_number, self.values = next(self.shredder)
            mpan_core_str = self.get_field(0, "MPAN Core")
            datum = {"mpan_core": parse_mpan_core(mpan_core_str)}
            channel_type_str = self.get_field(1, "Channel Type")
            datum["channel_type"] = parse_channel_type(channel_type_str)

            start_date_str = self.get_field(2, "Start Date")
            datum["start_date"] = validate_hh_start(
                to_utc(Datetime.strptime(start_date_str, "%Y-%m-%d %H:%M"))
            )

            value_str = self.get_field(3, "Value")
            datum["value"] = Decimal(value_str)

            status = self.get_field(4, "Status")
            if len(status) != 1:
                raise BadRequest(
                    "The status character must be one character in length."
                )
            datum["status"] = status
            return datum
        except BadRequest as e:
            e.description = "".join(
                "Problem at line number: ",
                str(self.line_number),
                ": ",
                str(self.values),
                ": ",
                e.description,
            )
            raise e

    def close(self):
        self.shredder.close()



import csv
from decimal import Decimal, InvalidOperation
from io import StringIO
from itertools import count

from werkzeug.exceptions import BadRequest

from zish import ZishLocationException, loads

from chellow.utils import parse_hh_start


def get_str(row, idx, name, lineno):
    return row[idx].strip()


def get_datetime(row, idx, name, lineno):
    val = get_str(row, idx, name, lineno)
    try:
        return parse_hh_start(val)
    except BadRequest as e:
        raise BadRequest(
            "Problem parsing the timestamp '"
            + val
            + "' at line number "
            + str(lineno)
            + " for the field '"
            + name
            + "' in column "
            + str(idx + 1)
            + ": "
            + e.description
        )


def get_decimal(row, idx, name, lineno):
    val = get_str(row, idx, name, lineno)
    try:
        return Decimal(val)
    except InvalidOperation as e:
        raise BadRequest(
            "Problem parsing the decimal '"
            + val
            + "' at line number "
            + str(lineno)
            + " for the field '"
            + name
            + "' at "
            + str(idx + 1)
            + ": "
            + str(e)
        )


class Parser:
    def __init__(self, file_bytes):
        self.csv_reader = csv.reader(
            StringIO(file_bytes.decode("utf8", errors="ignore")), skipinitialspace=True
        )
        self.titles = ",".join(next(self.csv_reader))
        self._line_number = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 2

    def _set_last_line(self, line):
        self.last_line = line
        return line

    def make_raw_bills(self):
        raw_bills = []
        for self._line_number, row in enumerate(self.csv_reader):
            bill_reference = row[0]
            if bill_reference == "" or bill_reference.startswith("#"):
                continue

            mprn = row[1]
            bill_type = row[2]
            account = row[3]
            issue_date = get_datetime(row, 4, "Issue Date", self.line_number)
            start_date = get_datetime(row, 5, "Start Date", self.line_number)
            finish_date = get_datetime(row, 6, "Finish Date", self.line_number)
            kwh = get_decimal(row, 7, "kWh", self.line_number)
            net = get_decimal(row, 8, "Net GBP", self.line_number)
            vat = get_decimal(row, 9, "VAT GBP", self.line_number)
            gross = get_decimal(row, 10, "Gross GBP", self.line_number)
            try:
                breakdown = loads(row[11])
            except ZishLocationException as e:
                raise BadRequest(
                    "Problem parsing the breakdown field at line number "
                    + str(self.line_number)
                    + ": "
                    + str(e)
                )

            reads = []

            for i in count(12, 10):
                if i > len(row) - 1 or len("".join(row[i:]).strip()) == 0:
                    break

                msn = get_str(row, i, "Meter Serial Number", self.line_number)
                unit = get_str(row, i + 1, "Unit", self.line_number).upper()
                correction_factor = get_decimal(
                    row, i + 2, "Correction Factor", self.line_number
                )

                calorific_value = get_str(
                    row, i + 3, "Calorific Value", self.line_number
                )
                if len(calorific_value) > 0:
                    calorific_value = get_decimal(
                        row, i + 3, "Calorific Value", self.line_number
                    )
                else:
                    calorific_value = None

                prev_date = get_datetime(row, i + 4, "Previous Date", self.line_number)
                prev_value = get_decimal(row, i + 5, "Previous Value", self.line_number)
                prev_type = get_str(row, i + 6, "Previous Type", self.line_number)
                pres_date = get_datetime(row, i + 7, "Previous Date", self.line_number)
                pres_value = get_decimal(row, i + 8, "Present Value", self.line_number)
                pres_type = get_str(row, i + 9, "Present Type", self.line_number)
                reads.append(
                    {
                        "msn": msn,
                        "unit": unit,
                        "correction_factor": correction_factor,
                        "calorific_value": calorific_value,
                        "prev_date": prev_date,
                        "prev_value": prev_value,
                        "prev_type_code": prev_type,
                        "pres_date": pres_date,
                        "pres_value": pres_value,
                        "pres_type_code": pres_type,
                    }
                )

            raw_bills.append(
                {
                    "mprn": mprn,
                    "reference": bill_reference,
                    "account": account,
                    "reads": reads,
                    "kwh": kwh,
                    "breakdown": breakdown,
                    "net_gbp": net,
                    "vat_gbp": vat,
                    "gross_gbp": gross,
                    "raw_lines": self.titles + "\n" + ",".join(row),
                    "bill_type_code": bill_type,
                    "start_date": start_date,
                    "finish_date": finish_date,
                    "issue_date": issue_date,
                }
            )
        return raw_bills



import csv
import traceback
from collections import defaultdict
from datetime import datetime as Datetime
from io import StringIO

from sqlalchemy import null, or_, true
from sqlalchemy.orm import joinedload

from werkzeug.exceptions import BadRequest

from chellow.e.computer import (
    SiteSource,
    SupplySource,
    contract_func,
    datum_range,
    displaced_era,
)
from chellow.models import Era, Llfc, MtcParticipant, Pc, SiteEra, Source, Supply
from chellow.utils import (
    c_months_u,
    hh_format,
    hh_max,
    hh_min,
    parse_hh_start,
    to_ct,
    to_utc,
)


def make_create_future_func_simple(contract_name, fnames=None):
    def create_future_func_simple(multiplier, constant):
        mult = float(multiplier)
        const = float(constant)
        if fnames is None:

            def future_func(ns):
                new_ns = {}
                for k, v in ns.items():
                    new_ns[k] = float(v) * mult + const

                return new_ns

        else:

            def future_func(ns):
                new_ns = {}
                for fname in fnames:
                    try:
                        new_ns[fname] = ns[fname] * mult + const
                    except KeyError:
                        raise BadRequest(
                            f"Can't find {fname} in rate script {ns} for contract name "
                            f"{contract_name}."
                        )

                return new_ns

        return future_func

    return create_future_func_simple


def make_create_future_func_monthly(contract_name, fnames):
    def create_future_func_monthly(multiplier, constant):
        def future_func(ns):
            new_ns = {}
            for fname in fnames:
                old_result = ns[fname]
                last_value = old_result[sorted(old_result.keys())[-1]]
                new_ns[fname] = defaultdict(
                    lambda: last_value,
                    [(k, v * multiplier + constant) for k, v in old_result.items()],
                )

            return new_ns

        return future_func

    return create_future_func_monthly


def make_site_deltas(
    sess, report_context, site, scenario_hh, forecast_from, supply_ids
):
    site_scenario_hh = scenario_hh.get(site.code, {})

    site_deltas = {"hhs": {}}
    delts = site_deltas["supply_deltas"] = {}
    for is_import in (True, False):
        delts[is_import] = {}
        for src in ("gen", "grid", "gen-grid", "3rd-party", "3rd-party-reverse", "sub"):
            delts[is_import][src] = {"site": {}}

    earliest_delta = to_utc(Datetime.max)
    latest_delta = to_utc(Datetime.min)

    found_hh = False
    for typ in ("used", "generated", "parasitic", "gen_grid"):
        hh_str = site_scenario_hh.get(typ, "")
        hh_data = site_scenario_hh[typ] = {}
        for row in csv.reader(StringIO(hh_str)):
            cells = [cell.strip() for cell in row]
            if len("".join(cells)) == 0:
                continue

            if len(cells) != 2:
                raise BadRequest(
                    f"Can't interpret the row {cells} it should be of the form "
                    f"'timestamp, kWh'"
                )

            date_str, kwh_str = cells
            ts = parse_hh_start(date_str)
            earliest_delta = min(ts, earliest_delta)
            latest_delta = max(ts, latest_delta)
            try:
                hh_data[ts] = float(kwh_str)
            except ValueError as e:
                raise BadRequest(
                    f"When looking at {typ} hh data, can't parse the kWh at {date_str} "
                    f": {e}"
                )
            found_hh = True

    if not found_hh:
        return site_deltas

    scenario_used = site_scenario_hh["used"]
    scenario_generated = site_scenario_hh["generated"]
    scenario_parasitic = site_scenario_hh["parasitic"]
    scenario_gen_grid = site_scenario_hh["gen_grid"]

    earliest_delta_ct = to_ct(earliest_delta)
    for month_start, month_finish in c_months_u(
        earliest_delta_ct.year, earliest_delta_ct.month, months=None
    ):
        if month_start > latest_delta:
            break
        chunk_start = hh_max(month_start, earliest_delta)
        chunk_finish = hh_min(month_finish, latest_delta)

        site_ds = SiteSource(
            sess, site, chunk_start, chunk_finish, forecast_from, report_context
        )
        hh_map = dict((h["start-date"], h) for h in site_ds.hh_data)

        for era in (
            sess.query(Era)
            .join(SiteEra)
            .join(Pc)
            .filter(
                SiteEra.site == site,
                SiteEra.is_physical == true(),
                Era.imp_mpan_core != null(),
                Pc.code != "00",
                Era.start_date <= chunk_finish,
                or_(Era.finish_date == null(), Era.finish_date >= chunk_start),
            )
        ):
            if supply_ids is not None and era.supply_id not in supply_ids:
                continue

            ss_start = hh_max(era.start_date, chunk_start)
            ss_finish = hh_min(era.finish_date, chunk_finish)

            ss = SupplySource(
                sess, ss_start, ss_finish, forecast_from, era, True, report_context
            )

            for hh in ss.hh_data:
                sdatum = hh_map[hh["start-date"]]
                sdatum["import-grid-kwh"] += hh["msp-kwh"]
                sdatum["used-kwh"] += hh["msp-kwh"]

        for era in (
            sess.query(Era)
            .join(SiteEra)
            .join(Pc)
            .join(Supply)
            .join(Source)
            .filter(
                SiteEra.site == site,
                SiteEra.is_physical == true(),
                Era.imp_mpan_core != null(),
                Era.start_date <= chunk_finish,
                or_(Era.finish_date == null(), Era.finish_date >= chunk_start),
                Source.code == "gen-grid",
            )
        ):
            if supply_ids is not None and era.supply_id not in supply_ids:
                continue

            ss_start = hh_max(era.start_date, chunk_start)
            ss_finish = hh_min(era.finish_date, chunk_finish)

            ss = SupplySource(
                sess, ss_start, ss_finish, forecast_from, era, False, report_context
            )

            for hh in ss.hh_data:
                sdatum = hh_map[hh["start-date"]]
                try:
                    sdatum["gen-grid-kwh"] += hh["msp-kwh"]
                except KeyError:
                    sdatum["gen-grid-kwh"] = hh["msp-kwh"]

        for hh_start, hh in hh_map.items():
            if hh_start in scenario_used:
                used_delt = scenario_used[hh_start] - hh["used-kwh"]
                imp_grid_delt = 0
                exp_grid_delt = 0

                if used_delt < 0:
                    diff = hh["import-grid-kwh"] + used_delt
                    if diff < 0:
                        imp_grid_delt -= hh["import-grid-kwh"]
                        exp_grid_delt -= diff
                    else:
                        imp_grid_delt += used_delt
                else:
                    diff = hh["export-grid-kwh"] - used_delt
                    if diff < 0:
                        exp_grid_delt -= hh["export-grid-kwh"]
                        imp_grid_delt -= diff
                    else:
                        exp_grid_delt -= used_delt

                try:
                    delts[False]["grid"]["site"][hh_start] += exp_grid_delt
                except KeyError:
                    delts[False]["grid"]["site"][hh_start] = exp_grid_delt

                try:
                    delts[True]["grid"]["site"][hh_start] += imp_grid_delt
                except KeyError:
                    delts[True]["grid"]["site"][hh_start] = imp_grid_delt

                hh["import-grid-kwh"] += imp_grid_delt
                hh["export-grid-kwh"] += exp_grid_delt
                hh["used-kwh"] += used_delt
                hh["msp-kwh"] -= exp_grid_delt

            if hh_start in scenario_generated:
                imp_gen_delt = scenario_generated[hh_start] - hh["import-gen-kwh"]
                imp_grid_delt = 0
                exp_grid_delt = 0

                if imp_gen_delt < 0:
                    diff = hh["export-grid-kwh"] + imp_gen_delt
                    if diff < 0:
                        exp_grid_delt -= hh["export-grid-kwh"]
                        imp_grid_delt -= diff
                    else:
                        exp_grid_delt += imp_gen_delt
                else:
                    diff = hh["import-grid-kwh"] - imp_gen_delt
                    if diff < 0:
                        imp_grid_delt -= hh["import-grid-kwh"]
                        exp_grid_delt -= diff
                    else:
                        imp_grid_delt -= imp_gen_delt

                try:
                    delts[True]["gen"]["site"][hh_start] += imp_gen_delt
                except KeyError:
                    delts[True]["gen"]["site"][hh_start] = imp_gen_delt

                try:
                    delts[False]["grid"]["site"][hh_start] += exp_grid_delt
                except KeyError:
                    delts[False]["grid"]["site"][hh_start] = exp_grid_delt

                try:
                    delts[True]["grid"]["site"][hh_start] += imp_grid_delt
                except KeyError:
                    delts[True]["grid"]["site"][hh_start] = imp_grid_delt

                hh["import-grid-kwh"] += imp_grid_delt
                hh["export-grid-kwh"] += exp_grid_delt
                hh["import-gen-kwh"] += imp_gen_delt
                hh["msp-kwh"] -= imp_grid_delt

            if hh_start in scenario_parasitic:
                exp_gen_delt = scenario_parasitic[hh_start] - hh["export-gen-kwh"]
                imp_grid_delt = 0
                exp_grid_delt = 0

                if exp_gen_delt < 0:
                    diff = hh["import-grid-kwh"] + exp_gen_delt
                    if diff < 0:
                        imp_grid_delt -= hh["import-grid-kwh"]
                        exp_grid_delt -= diff
                    else:
                        imp_grid_delt += exp_gen_delt
                else:
                    diff = hh["export-grid-kwh"] - exp_gen_delt
                    if diff < 0:
                        exp_grid_delt -= hh["export-grid-kwh"]
                        imp_grid_delt -= diff
                    else:
                        exp_grid_delt -= exp_gen_delt

                try:
                    delts[False]["gen"]["site"][hh_start] += imp_gen_delt
                except KeyError:
                    delts[False]["gen"]["site"][hh_start] = exp_gen_delt

                try:
                    delts[False]["grid"]["site"][hh_start] += exp_grid_delt
                except KeyError:
                    delts[False]["grid"]["site"][hh_start] = exp_grid_delt

                try:
                    delts[True]["grid"]["site"][hh_start] += imp_grid_delt
                except KeyError:
                    delts[True]["grid"]["site"][hh_start] = imp_grid_delt

                hh["import-grid-kwh"] += imp_grid_delt
                hh["export-grid-kwh"] += exp_grid_delt
                hh["export-gen-kwh"] += exp_gen_delt
                hh["msp-kwh"] -= imp_grid_delt

            if hh_start in scenario_gen_grid:
                gen_grid_delt = scenario_gen_grid[hh_start] - hh["gen-grid-kwh"]

                try:
                    delts[False]["gen-grid"]["site"][hh_start] += gen_grid_delt
                except KeyError:
                    delts[False]["gen-grid"]["site"][hh_start] = gen_grid_delt

                hh["import-gen-kwh"] += gen_grid_delt
                hh["export-grid-kwh"] += gen_grid_delt

            site_deltas["hhs"][hh_start] = hh

    sup_deltas = site_deltas["supply_deltas"][False]["grid"]["site"]
    if all(v == 0 for v in sup_deltas.values()):
        sup_deltas.clear()

    return site_deltas


def make_calcs(
    sess,
    site,
    start_date,
    finish_date,
    supply_ids,
    site_deltas,
    forecast_from,
    report_context,
    era_maps,
    data_source_bill,
):
    meter_order = {"hh": 0, "amr": 1, "nhh": 2, "unmetered": 3}
    site_gen_types = set()
    calcs = []
    for era in (
        sess.query(Era)
        .join(SiteEra)
        .join(Pc)
        .filter(
            SiteEra.site == site,
            SiteEra.is_physical == true(),
            Era.start_date <= finish_date,
            or_(Era.finish_date == null(), Era.finish_date >= start_date),
        )
        .options(
            joinedload(Era.ssc),
            joinedload(Era.dc_contract),
            joinedload(Era.mop_contract),
            joinedload(Era.imp_supplier_contract),
            joinedload(Era.exp_supplier_contract),
            joinedload(Era.channels),
            joinedload(Era.imp_llfc).joinedload(Llfc.voltage_level),
            joinedload(Era.exp_llfc).joinedload(Llfc.voltage_level),
            joinedload(Era.cop),
            joinedload(Era.supply).joinedload(Supply.dno),
            joinedload(Era.supply).joinedload(Supply.gsp_group),
            joinedload(Era.supply).joinedload(Supply.source),
            joinedload(Era.mtc_participant).joinedload(MtcParticipant.meter_type),
            joinedload(Era.pc),
            joinedload(Era.energisation_status),
            joinedload(Era.site_eras),
        )
        .order_by(Era.supply_id, Era.start_date)
    ):
        supply = era.supply
        if data_source_bill is not None and supply.dno.dno_code in ("88", "99"):
            continue

        if supply.generator_type is not None:
            site_gen_types.add(supply.generator_type.code)

        if supply_ids is not None and supply.id not in supply_ids:
            continue

        ss_start = hh_max(era.start_date, start_date)
        ss_finish = hh_min(era.finish_date, finish_date)

        if era.imp_mpan_core is None:
            imp_ss = None
        else:
            sup_deltas = site_deltas["supply_deltas"][True][supply.source.code]
            imp_ss = SupplySource(
                sess,
                ss_start,
                ss_finish,
                forecast_from,
                era,
                True,
                report_context,
                era_maps=era_maps,
                deltas=sup_deltas,
                bill=data_source_bill if era.pc.code == "00" else None,
            )
            imp_supplier_contract = imp_ss.supplier_contract
            if imp_supplier_contract is not None:
                import_vb_function = contract_func(
                    report_context, imp_supplier_contract, "virtual_bill"
                )
                if import_vb_function is None:
                    raise BadRequest(
                        f"The supplier contract {imp_supplier_contract.name} "
                        " doesn't have the virtual_bill() function."
                    )
                try:
                    import_vb_function(imp_ss)
                except (AttributeError, TypeError) as e:
                    raise BadRequest(
                        f"Problem with virtual bill of supplier contract "
                        f"{imp_supplier_contract.id} {e} {traceback.format_exc()}"
                    )
                except BadRequest as e:
                    raise BadRequest(
                        f"{e.description} Problem with virtual bill of supplier "
                        f"contract {imp_supplier_contract.id} {e}"
                    )

        if era.exp_mpan_core is None:
            exp_ss = None
            measurement_type = imp_ss.measurement_type
        else:
            sup_deltas = site_deltas["supply_deltas"][False][supply.source.code]

            exp_ss = SupplySource(
                sess,
                ss_start,
                ss_finish,
                forecast_from,
                era,
                False,
                report_context,
                era_maps=era_maps,
                deltas=sup_deltas,
                bill=data_source_bill if era.pc.code == "00" else None,
            )
            measurement_type = exp_ss.measurement_type
            exp_supplier_contract = exp_ss.supplier_contract
            if exp_supplier_contract is not None:
                export_vb_function = contract_func(
                    report_context, exp_supplier_contract, "virtual_bill"
                )
                try:
                    export_vb_function(exp_ss)
                except (AttributeError, TypeError) as e:
                    raise BadRequest(
                        f"Problem with virtual bill of supplier contract "
                        f"{exp_supplier_contract.id} {e} {traceback.format_exc()}"
                    )
                except BadRequest as e:
                    raise BadRequest(
                        f"{e.description} Problem with virtual bill of supplier "
                        f"contract {exp_supplier_contract.id}"
                    )

        order = (
            f"{meter_order[measurement_type]}_{supply.id}_{hh_format(era.start_date)}"
        )
        calcs.append((order, era.imp_mpan_core, era.exp_mpan_core, imp_ss, exp_ss))

    start_date_ct, finish_date_ct = to_ct(start_date), to_ct(finish_date)
    for month_start, month_finish in c_months_u(
        start_year=start_date_ct.year,
        start_month=start_date_ct.month,
        finish_year=finish_date_ct.year,
        finish_month=finish_date_ct.month,
    ):
        ss_start = hh_max(month_start, start_date)
        ss_finish = hh_min(month_finish, finish_date)
        # Check if gen deltas haven't been consumed
        extra_sss = set()
        for is_imp in (True, False):
            sup_deltas = site_deltas["supply_deltas"][is_imp]["gen"]["site"]
            if len(list(t for t in sup_deltas if ss_start <= t <= ss_finish)) > 0:
                extra_sss.add(is_imp)

        disp_era = displaced_era(
            sess,
            report_context,
            site,
            ss_start,
            ss_finish,
            forecast_from,
            has_scenario_generation=len(extra_sss) > 0,
        )

        if disp_era is not None and supply_ids is None:
            site_ds = SiteSource(
                sess,
                site,
                ss_start,
                ss_finish,
                forecast_from,
                report_context,
                disp_era,
                era_maps=era_maps,
                deltas=site_deltas,
                bill=data_source_bill,
            )

            disp_supplier_contract = disp_era.imp_supplier_contract
            disp_vb_function = contract_func(
                report_context, disp_supplier_contract, "displaced_virtual_bill"
            )
            if disp_vb_function is None:
                raise BadRequest(
                    f"The supplier contract {disp_supplier_contract.name} "
                    f" doesn't have the displaced_virtual_bill() function."
                )
            disp_vb_function(site_ds)

            calcs.append((f"1_{hh_format(ss_start)}", "displaced", None, site_ds, None))

        if len(extra_sss) > 0:
            if True in extra_sss:
                sup_deltas = site_deltas["supply_deltas"][True]["gen"]
                imp_ss_name = site.code + "_extra_gen_TRUE"
                imp_ss = ScenarioSource(
                    sess,
                    ss_start,
                    ss_finish,
                    True,
                    report_context,
                    sup_deltas,
                    disp_era.imp_supplier_contract,
                    imp_ss_name,
                    bill=data_source_bill,
                )
            else:
                imp_ss_name = imp_ss = None
            if False in extra_sss:
                exp_ss_name = site.code + "_extra_gen_FALSE"
                sup_deltas = site_deltas["supply_deltas"][False]["gen"]
                exp_ss = ScenarioSource(
                    sess,
                    ss_start,
                    ss_finish,
                    False,
                    report_context,
                    sup_deltas,
                    disp_era.imp_supplier_contract,
                    imp_ss_name,
                    bill=data_source_bill,
                )
            else:
                exp_ss_name = exp_ss = None

            calcs.append(("0", imp_ss_name, exp_ss_name, imp_ss, exp_ss))

        # Check if exp grid deltas haven't been consumed
        sup_deltas = site_deltas["supply_deltas"][False]["grid"]
        if len(list(t for t in sup_deltas["site"] if ss_start <= t <= ss_finish)) > 0:
            ss_name = site.code + "_extra_grid_export"
            ss = SupplySource(
                sess,
                ss_start,
                ss_finish,
                forecast_from,
                disp_era,
                False,
                report_context,
                era_maps=era_maps,
                deltas=sup_deltas,
                bill=data_source_bill,
            )

            calcs.append(("0", None, ss_name, None, ss))
    return calcs, site_gen_types


class ScenarioSource:
    def __init__(
        self,
        sess,
        start_date,
        finish_date,
        is_import,
        caches,
        deltas,
        supplier_contract,
        mpan_core,
        bill=None,
    ):
        self.sess = sess
        self.supply = None
        self.mpan_core = mpan_core
        self.supply_name = mpan_core
        self.start_date = start_date
        self.is_import = is_import
        self.caches = caches
        self.deltas = deltas
        self.years_back = 0
        self.source_code = "gen"
        self.dno_code = "99"
        self.llfc_code = "510"
        self.voltage_level_code = "LV"
        self.is_substation = False
        self.gsp_group_code = "_L"
        self.supplier_bill = {"net-gbp": 0}
        self.dc_bill = {"net-gbp": 0}
        self.mop_bill = {"net-gbp": 0}
        self.supplier_contract = None
        self.dc_contract = None
        self.mop_contract = None
        self.supplier_rate_sets = defaultdict(set)
        self.is_displaced = False
        self.sc = 0
        self.pc_code = "00"
        self.mop_rate_sets = defaultdict(set)
        self.dc_rate_sets = defaultdict(set)
        self.generator_type_code = "chp"
        self.msn = ""
        self.measurement_type = "hh"
        self.bill = bill
        self.hh_data = list(
            d.copy()
            for d in datum_range(
                sess, self.caches, self.years_back, start_date, finish_date
            )
        )
        if self.deltas is not None:
            site_deltas = self.deltas["site"]

            try:
                sup_deltas = self.deltas[self.mpan_core]
            except KeyError:
                sup_deltas = self.deltas[self.mpan_core] = {}

            for hh in self.hh_data:
                hh_start = hh["start-date"]
                if hh_start in sup_deltas:
                    delt = sup_deltas[hh_start]
                elif hh_start in site_deltas:
                    delt = sup_deltas[hh_start] = site_deltas[hh_start]
                    del site_deltas[hh_start]
                else:
                    continue

                hh["msp-kwh"] += delt
                hh["msp-kw"] += delt * 2



import csv
import sys
import threading
import traceback

from flask import g

from sqlalchemy import or_, select
from sqlalchemy.sql.expression import null, true

from chellow.dloads import open_file
from chellow.e.computer import SupplySource, forecast_date
from chellow.models import Era, RSession, Site, SiteEra, Supply
from chellow.utils import (
    csv_make_val,
    hh_format,
    hh_range,
    req_date,
    req_int,
)
from chellow.views import chellow_redirect


def content(supply_id, start_date, finish_date, user):
    caches = {}
    f = None
    try:
        with RSession() as sess:
            supply = Supply.get_by_id(sess, supply_id)

            f_date = forecast_date()

            f = open_file(
                f"supply_virtual_bills_hh_{supply_id}.csv", user, mode="w", newline=""
            )
            w = csv.writer(f, lineterminator="\n")

            mop_titles = []
            dc_titles = []
            imp_supplier_titles = []
            exp_supplier_titles = []

            for era in sess.execute(
                select(Era).where(
                    Era.supply == supply,
                    Era.start_date <= finish_date,
                    or_(Era.finish_date == null(), Era.finish_date >= start_date),
                )
            ).scalars():
                ds = SupplySource(
                    sess, era.start_date, era.start_date, f_date, era, True, caches
                )

                for t in ds.contract_func(era.mop_contract, "virtual_bill_titles")():
                    if t not in mop_titles:
                        mop_titles.append(t)

                for t in ds.contract_func(era.dc_contract, "virtual_bill_titles")():
                    if t not in dc_titles:
                        dc_titles.append(t)

                if era.imp_supplier_contract is not None:
                    for t in ds.contract_func(
                        era.imp_supplier_contract, "virtual_bill_titles"
                    )():
                        if t not in imp_supplier_titles:
                            imp_supplier_titles.append(t)

                if era.exp_supplier_contract is not None:
                    ds = SupplySource(
                        sess,
                        era.start_date,
                        era.start_date,
                        f_date,
                        era,
                        False,
                        caches,
                    )
                    for t in ds.contract_func(
                        era.exp_supplier_contract, "virtual_bill_titles"
                    )():
                        if t not in exp_supplier_titles:
                            exp_supplier_titles.append(t)

            titles = [
                "mpan_core",
                "site_code",
                "site_name",
                "hh_start",
            ]
            for pref, t in (
                ("mop", mop_titles),
                ("dc", dc_titles),
                ("imp_supplier", imp_supplier_titles),
                ("exp_supplier", exp_supplier_titles),
            ):
                titles.append("")
                titles.extend([f"{pref}_{n}" for n in t])

            w.writerow(titles)

            for hh_start in hh_range(caches, start_date, finish_date):
                era = sess.execute(
                    select(Era).where(
                        Era.supply == supply,
                        Era.start_date <= hh_start,
                        or_(Era.finish_date == null(), Era.finish_date >= hh_start),
                    )
                ).scalar_one()

                site = sess.execute(
                    select(Site)
                    .join(SiteEra)
                    .where(SiteEra.era == era, SiteEra.is_physical == true())
                ).scalar_one()

                ds = SupplySource(sess, hh_start, hh_start, f_date, era, True, caches)

                vals = {
                    "mpan_core": ds.mpan_core,
                    "site_code": site.code,
                    "site_name": site.name,
                    "hh_start": hh_format(ds.start_date),
                }

                ds.contract_func(era.mop_contract, "virtual_bill")(ds)
                for k, v in ds.mop_bill.items():
                    vals[f"mop_{k}"] = v

                ds.contract_func(era.dc_contract, "virtual_bill")(ds)
                for k, v in ds.dc_bill.items():
                    vals[f"dc_{k}"] = v

                if era.imp_supplier_contract is not None:
                    ds.contract_func(era.imp_supplier_contract, "virtual_bill")(ds)
                    for k, v in ds.supplier_bill.items():
                        vals[f"imp_supplier_{k}"] = v

                if era.exp_supplier_contract is not None:
                    ds = SupplySource(
                        sess, hh_start, hh_start, f_date, era, False, caches
                    )
                    ds.contract_func(era.exp_supplier_contract, "virtual_bill")(ds)
                    for k, v in ds.supplier_bill.items():
                        vals[f"exp_supplier_{k}"] = v

                w.writerow([csv_make_val(vals.get(t)) for t in titles])
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        if f is not None:
            f.write(msg)
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    supply_id = req_int("supply_id")
    start_date = req_date("start")
    finish_date = req_date("finish")

    args = supply_id, start_date, finish_date, g.user
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import csv
import sys
import threading
import traceback
from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal
from itertools import combinations

from dateutil.relativedelta import relativedelta

from flask import g, request

from sqlalchemy import or_, select
from sqlalchemy.orm import joinedload, subqueryload
from sqlalchemy.sql.expression import null, true

from werkzeug.exceptions import BadRequest

from zish import ZishLocationException, loads

from chellow.dloads import open_file
from chellow.e.computer import SupplySource, contract_func
from chellow.models import (
    Batch,
    Bill,
    Contract,
    Era,
    Llfc,
    MtcParticipant,
    RSession,
    RegisterRead,
    ReportRun,
    Site,
    SiteEra,
    Supply,
    User,
)
from chellow.utils import (
    HH,
    csv_make_val,
    hh_format,
    hh_max,
    hh_min,
    hh_range,
    parse_mpan_core,
    req_date,
    req_int,
    req_str,
    to_utc,
)
from chellow.views import chellow_redirect


def add_gap(caches, gaps, elem, start_date, finish_date, is_virtual, gbp):
    try:
        elgap = gaps[elem]
    except KeyError:
        elgap = gaps[elem] = {}

    hhs = hh_range(caches, start_date, finish_date)
    hhgbp = 0 if gbp is None else gbp / len(hhs)

    for hh_start in hhs:
        try:
            hhgap = elgap[hh_start]
        except KeyError:
            hhgap = elgap[hh_start] = {
                "has_covered": False,
                "has_virtual": False,
                "gbp": 0,
            }

        if is_virtual:
            hhgap["has_virtual"] = True
            hhgap["gbp"] = hhgbp
        else:
            hhgap["has_covered"] = True


def find_elements(bill):
    try:
        keys = [k for k in loads(bill.breakdown).keys() if k.endswith("-gbp")]
        return set(k[:-4] for k in keys)
    except ZishLocationException as e:
        raise BadRequest(
            f"Can't parse the breakdown for bill id {bill.id} attached to batch id "
            f"{bill.batch.id}: {e}"
        )


def content(
    batch_id,
    bill_id,
    contract_id,
    start_date,
    finish_date,
    user_id,
    mpan_cores,
    fname_additional,
    report_run_id,
):
    caches = {}
    tmp_file = sess = supply_id = None
    forecast_date = to_utc(Datetime.max)

    try:
        with RSession() as sess:
            user = User.get_by_id(sess, user_id)
            tmp_file = open_file(
                f"bill_check_{fname_additional}.csv", user, mode="w", newline=""
            )
            writer = csv.writer(tmp_file, lineterminator="\n")

            bills = (
                sess.query(Bill)
                .order_by(Bill.supply_id, Bill.reference)
                .options(
                    joinedload(Bill.supply),
                    subqueryload(Bill.reads).joinedload(RegisterRead.present_type),
                    subqueryload(Bill.reads).joinedload(RegisterRead.previous_type),
                    joinedload(Bill.batch),
                )
            )

            if len(mpan_cores) > 0:
                mpan_cores = list(map(parse_mpan_core, mpan_cores))
                supply_ids = [
                    i[0]
                    for i in sess.query(Era.supply_id)
                    .filter(
                        or_(
                            Era.imp_mpan_core.in_(mpan_cores),
                            Era.exp_mpan_core.in_(mpan_cores),
                        )
                    )
                    .distinct()
                ]
                bills = bills.join(Supply).filter(Supply.id.in_(supply_ids))

            if batch_id is not None:
                batch = Batch.get_by_id(sess, batch_id)
                bills = bills.filter(Bill.batch == batch)
                contract = batch.contract
            elif bill_id is not None:
                bill = Bill.get_by_id(sess, bill_id)
                bills = bills.filter(Bill.id == bill.id)
                contract = bill.batch.contract
            elif contract_id is not None:
                contract = Contract.get_by_id(sess, contract_id)
                bills = bills.join(Batch).filter(
                    Batch.contract == contract,
                    Bill.start_date <= finish_date,
                    Bill.finish_date >= start_date,
                )

            vbf = contract_func(caches, contract, "virtual_bill")
            if vbf is None:
                raise BadRequest(
                    f"The contract {contract.name} doesn't have a function "
                    f"virtual_bill."
                )

            virtual_bill_titles_func = contract_func(
                caches, contract, "virtual_bill_titles"
            )
            if virtual_bill_titles_func is None:
                raise BadRequest(
                    f"The contract {contract.name} doesn't have a function "
                    f"virtual_bill_titles."
                )
            virtual_bill_titles = virtual_bill_titles_func()

            titles = [
                "batch",
                "bill-reference",
                "bill-type",
                "bill-kwh",
                "bill-net-gbp",
                "bill-vat-gbp",
                "bill-start-date",
                "bill-finish-date",
                "imp-mpan-core",
                "exp-mpan-core",
                "site-code",
                "site-name",
                "covered-from",
                "covered-to",
                "covered-bills",
                "metered-kwh",
            ]
            for t in virtual_bill_titles:
                titles.append("covered-" + t)
                titles.append("virtual-" + t)
                if t.endswith("-gbp"):
                    titles.append("difference-" + t)

            writer.writerow(titles)

            bill_map = defaultdict(set, {})
            for bill in bills:
                bill_map[bill.supply.id].add(bill.id)

            for supply_id, bill_ids in bill_map.items():
                _process_supply(
                    sess,
                    caches,
                    supply_id,
                    bill_ids,
                    forecast_date,
                    contract,
                    vbf,
                    virtual_bill_titles,
                    writer,
                    titles,
                    report_run_id,
                )
        ReportRun.w_update(report_run_id, "finished")

    except BadRequest as e:
        if supply_id is None:
            prefix = "Problem: "
        else:
            prefix = f"Problem with supply {supply_id}:"
        tmp_file.write(prefix + e.description)
        ReportRun.w_update(report_run_id, "problem")
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        if supply_id is None:
            prefix = "Problem: "
        else:
            prefix = f"Problem with supply {supply_id}:"
        tmp_file.write(prefix + msg)
        ReportRun.w_update(report_run_id, "interrupted")
    finally:
        if tmp_file is not None:
            tmp_file.close()


def do_get(sess):
    return do_post(sess)


def do_post(sess):
    batch_id = bill_id = contract_id = start_date = finish_date = None
    if "mpan_cores" in request.values:
        mpan_cores = req_str("mpan_cores").splitlines()
    else:
        mpan_cores = []

    fname_additional = ""

    if "batch_id" in request.values:
        batch_id = req_int("batch_id")
        batch = Batch.get_by_id(sess, batch_id)
        fname_additional = f"_batch_{batch.reference}"
    elif "bill_id" in request.values:
        bill_id = req_int("bill_id")
        bill = Bill.get_by_id(sess, bill_id)
        fname_additional = "bill_" + str(bill.id)
    elif "contract_id" in request.values:
        contract_id = req_int("contract_id")
        contract = Contract.get_by_id(sess, contract_id)

        start_date = req_date("start_date")
        finish_date = req_date("finish_date")

        s = ["contract", str(contract.id)]
        for dt in (start_date, finish_date):
            s.append(hh_format(dt).replace(" ", "T").replace(":", ""))
        fname_additional = "_".join(s)
    else:
        raise BadRequest(
            "The bill check needs a batch_id, a bill_id or a start_date and "
            "finish_date."
        )

    report_run = ReportRun.insert(
        sess,
        "bill_check",
        g.user,
        fname_additional,
        {},
    )
    sess.commit()

    args = (
        batch_id,
        bill_id,
        contract_id,
        start_date,
        finish_date,
        g.user.id,
        mpan_cores,
        fname_additional,
        report_run.id,
    )
    threading.Thread(target=content, args=args).start()
    return chellow_redirect(f"/report_runs/{report_run.id}", 303)


def _process_supply(
    sess,
    caches,
    supply_id,
    bill_ids,
    forecast_date,
    contract,
    vbf,
    virtual_bill_titles,
    writer,
    titles,
    report_run_id,
):
    gaps = {}
    data_sources = {}
    market_role_code = contract.market_role.code

    while len(bill_ids) > 0:
        bill_id = list(sorted(bill_ids))[0]
        bill_ids.remove(bill_id)
        bill = (
            sess.query(Bill)
            .filter(Bill.id == bill_id)
            .options(
                joinedload(Bill.batch),
                joinedload(Bill.bill_type),
                joinedload(Bill.reads),
                joinedload(Bill.supply),
                joinedload(Bill.reads).joinedload(RegisterRead.present_type),
                joinedload(Bill.reads).joinedload(RegisterRead.previous_type),
            )
            .one()
        )
        virtual_bill = {"problem": ""}
        supply = bill.supply

        read_dict = {}
        for read in bill.reads:
            gen_start = read.present_date.replace(hour=0).replace(minute=0)
            gen_finish = gen_start + relativedelta(days=1) - HH
            msn_match = False
            read_msn = read.msn
            for read_era in supply.find_eras(sess, gen_start, gen_finish):
                if read_msn == read_era.msn:
                    msn_match = True
                    break

            if not msn_match:
                virtual_bill["problem"] += (
                    f"The MSN {read_msn} of the register read {read.id} doesn't match "
                    f"the MSN of the era."
                )

            for dt, typ in [
                (read.present_date, read.present_type),
                (read.previous_date, read.previous_type),
            ]:
                key = str(dt) + "-" + read.msn
                try:
                    if typ != read_dict[key]:
                        virtual_bill[
                            "problem"
                        ] += f" Reads taken on {dt} have differing read types."
                except KeyError:
                    read_dict[key] = typ

        bill_start = bill.start_date
        bill_finish = bill.finish_date

        covered_start = bill_start
        covered_finish = bill_start
        covered_bdown = {"sum-msp-kwh": 0, "net-gbp": 0, "vat-gbp": 0, "problem": ""}

        vb_elems = set()
        enlarged = True

        while enlarged:
            enlarged = False
            covered_elems = find_elements(bill)
            covered_bills = dict(
                (b.id, b)
                for b in sess.scalars(
                    select(Bill)
                    .join(Batch)
                    .where(
                        Bill.supply == supply,
                        Bill.start_date <= covered_finish,
                        Bill.finish_date >= covered_start,
                        Batch.contract == contract,
                    )
                    .order_by(Bill.start_date, Bill.issue_date)
                )
            )
            while True:
                to_del = None
                for a, b in combinations(covered_bills.values(), 2):
                    if all(
                        (
                            a.start_date == b.start_date,
                            a.finish_date == b.finish_date,
                            a.net == -1 * b.net,
                            a.vat == -1 * b.vat,
                            a.gross == -1 * b.gross,
                        )
                    ):
                        to_del = (a.id, b.id)
                        break
                if to_del is None:
                    break
                else:
                    for k in to_del:
                        del covered_bills[k]
                        bill_ids.discard(k)

            for k, covered_bill in tuple(covered_bills.items()):
                elems = find_elements(covered_bill)
                if elems.isdisjoint(covered_elems):
                    if k != bill.id:
                        del covered_bills[k]
                        continue
                else:
                    covered_elems.update(elems)

                if covered_bill.start_date < covered_start:
                    covered_start = covered_bill.start_date
                    enlarged = True
                    break

                if covered_bill.finish_date > covered_finish:
                    covered_finish = covered_bill.finish_date
                    enlarged = True
                    break

        if len(covered_bills) == 0:
            continue

        primary_covered_bill = None
        for covered_bill in covered_bills.values():
            bill_ids.discard(covered_bill.id)
            covered_bdown["net-gbp"] += float(covered_bill.net)
            covered_bdown["vat-gbp"] += float(covered_bill.vat)
            covered_bdown["sum-msp-kwh"] += float(covered_bill.kwh)
            for k, v in loads(covered_bill.breakdown).items():
                if k in ("raw_lines", "raw-lines", "vat"):
                    continue

                if isinstance(v, list):
                    try:
                        covered_bdown[k].update(set(v))
                    except KeyError:
                        covered_bdown[k] = set(v)
                    except AttributeError as e:
                        raise BadRequest(
                            f"For key {k} in {[b.id for b in covered_bills.values()]} "
                            f"the value {v} can't be added to the existing value "
                            f"{covered_bdown[k]}. {e}"
                        )
                else:
                    if isinstance(v, Decimal):
                        v = float(v)
                    try:
                        covered_bdown[k] += v
                    except KeyError:
                        covered_bdown[k] = v
                    except TypeError as detail:
                        raise BadRequest(
                            f"For key {k} in {[b.id for b in covered_bills.values()]} "
                            f"the value {v} can't be added to the existing value "
                            f"{covered_bdown[k]}. {detail}"
                        )

                    if k.endswith("-gbp"):
                        elem = k[:-4]
                        covered_elems.add(elem)
                        add_gap(
                            caches,
                            gaps,
                            elem,
                            covered_bill.start_date,
                            covered_bill.finish_date,
                            False,
                            v,
                        )

            if primary_covered_bill is None or (
                (covered_bill.finish_date - covered_bill.start_date)
                > (primary_covered_bill.finish_date - primary_covered_bill.start_date)
            ):
                primary_covered_bill = covered_bill

        metered_kwh = 0
        for era in (
            sess.query(Era)
            .filter(
                Era.supply == supply,
                Era.start_date <= covered_finish,
                or_(Era.finish_date == null(), Era.finish_date >= covered_start),
            )
            .distinct()
            .options(
                joinedload(Era.channels),
                joinedload(Era.cop),
                joinedload(Era.dc_contract),
                joinedload(Era.exp_llfc),
                joinedload(Era.exp_llfc).joinedload(Llfc.voltage_level),
                joinedload(Era.exp_supplier_contract),
                joinedload(Era.imp_llfc),
                joinedload(Era.imp_llfc).joinedload(Llfc.voltage_level),
                joinedload(Era.imp_supplier_contract),
                joinedload(Era.mop_contract),
                joinedload(Era.mtc_participant).joinedload(MtcParticipant.meter_type),
                joinedload(Era.pc),
                joinedload(Era.supply).joinedload(Supply.dno),
                joinedload(Era.supply).joinedload(Supply.gsp_group),
                joinedload(Era.supply).joinedload(Supply.source),
            )
        ):
            chunk_start = hh_max(covered_start, era.start_date)
            chunk_finish = hh_min(covered_finish, era.finish_date)

            if contract not in (
                era.mop_contract,
                era.dc_contract,
                era.imp_supplier_contract,
                era.exp_supplier_contract,
            ):
                virtual_bill["problem"] += (
                    f"From {hh_format(chunk_start)} to {hh_format(chunk_finish)} "
                    f"the contract of the era doesn't match the contract of the bill."
                )
                continue

            if contract.market_role.code == "X":
                polarity = contract != era.exp_supplier_contract
            else:
                polarity = era.imp_supplier_contract is not None

            try:
                ds_key = (
                    chunk_start,
                    chunk_finish,
                    forecast_date,
                    era.id,
                    polarity,
                    primary_covered_bill.id,
                )
                data_source = data_sources[ds_key]
            except KeyError:
                data_source = data_sources[ds_key] = SupplySource(
                    sess,
                    chunk_start,
                    chunk_finish,
                    forecast_date,
                    era,
                    polarity,
                    caches,
                    primary_covered_bill,
                )
                vbf(data_source)

            if data_source.measurement_type == "hh":
                metered_kwh += sum(h["msp-kwh"] for h in data_source.hh_data)
            else:
                ds = SupplySource(
                    sess,
                    chunk_start,
                    chunk_finish,
                    forecast_date,
                    era,
                    polarity,
                    caches,
                )
                metered_kwh += sum(h["msp-kwh"] for h in ds.hh_data)

            if market_role_code == "X":
                vb = data_source.supplier_bill
                vb_hhs = data_source.supplier_bill_hhs
            elif market_role_code == "C":
                vb = data_source.dc_bill
                vb_hhs = data_source.dc_bill_hhs
            elif market_role_code == "M":
                vb = data_source.mop_bill
                vb_hhs = data_source.mop_bill_hhs
            else:
                raise BadRequest("Odd market role.")

            for k, v in vb.items():
                try:
                    if isinstance(v, set):
                        virtual_bill[k].update(v)
                    else:
                        virtual_bill[k] += v
                except KeyError:
                    virtual_bill[k] = v
                except TypeError as detail:
                    raise BadRequest(f"For key {k} and value {v}. {detail}")

            for dt, bl in vb_hhs.items():
                for k, v in bl.items():
                    if all((k.endswith("-gbp"), k != "net-gbp", v != 0)):
                        add_gap(caches, gaps, k[:-4], dt, dt, True, v)

            for k in virtual_bill.keys():
                if k.endswith("-gbp"):
                    vb_elems.add(k[:-4])

        long_map = {}
        vb_keys = set(virtual_bill.keys())
        for elem in sorted(vb_elems, key=len, reverse=True):
            els = long_map[elem] = set()
            for k in tuple(vb_keys):
                if k.startswith(elem + "-"):
                    els.add(k)
                    vb_keys.remove(k)

        for elem in vb_elems.difference(covered_elems):
            for k in long_map[elem]:
                del virtual_bill[k]

        for elem in covered_elems.difference(vb_elems):
            covered_bdown["problem"] += (
                f"The element {elem} is in the covered bills, but not in the "
                f"virtual bill. "
            )

        try:
            del virtual_bill["net-gbp"]
        except KeyError:
            pass

        virtual_bill["net-gbp"] = sum(
            v for k, v in virtual_bill.items() if k.endswith("-gbp")
        )

        era = supply.find_era_at(sess, bill_finish)
        if era is None:
            imp_mpan_core = exp_mpan_core = None
            site_code = site_name = None
            virtual_bill["problem"] += "This bill finishes before or after the supply. "
        else:
            imp_mpan_core = era.imp_mpan_core
            exp_mpan_core = era.exp_mpan_core

            site = (
                sess.query(Site)
                .join(SiteEra)
                .filter(SiteEra.is_physical == true(), SiteEra.era == era)
                .one()
            )
            site_code = site.code
            site_name = site.name

        # Find bill to use for header data
        if bill.id not in covered_bills:
            for cbill in covered_bills.values():
                if bill.batch == cbill.batch:
                    bill = cbill

        values = {
            "batch": bill.batch.reference,
            "bill-reference": bill.reference,
            "bill-type": bill.bill_type.code,
            "bill-kwh": bill.kwh,
            "bill-net-gbp": bill.net,
            "bill-vat-gbp": bill.vat,
            "bill-start-date": bill_start,
            "bill-finish-date": bill_finish,
            "imp-mpan-core": imp_mpan_core,
            "exp-mpan-core": exp_mpan_core,
            "site-code": site_code,
            "site-name": site_name,
            "covered-from": covered_start,
            "covered-to": covered_finish,
            "covered-bills": sorted(covered_bills.keys()),
            "metered-kwh": metered_kwh,
        }
        for title in virtual_bill_titles:
            try:
                cov_val = covered_bdown[title]
                del covered_bdown[title]
            except KeyError:
                cov_val = None

            values[f"covered-{title}"] = cov_val

            try:
                virt_val = virtual_bill[title]
                del virtual_bill[title]
            except KeyError:
                virt_val = None

            values[f"virtual-{title}"] = virt_val

            if title.endswith("-gbp"):
                if isinstance(virt_val, (int, float, Decimal)):
                    if isinstance(cov_val, (int, float, Decimal)):
                        diff_val = float(cov_val) - float(virt_val)
                    else:
                        diff_val = 0 - float(virt_val)
                else:
                    diff_val = 0

                values[f"difference-{title}"] = diff_val

        report_run_titles = list(titles)
        for title in sorted(virtual_bill.keys()):
            virt_val = virtual_bill[title]
            virt_title = f"virtual-{title}"
            values[virt_title] = virt_val
            report_run_titles.append(virt_title)
            if title in covered_bdown:
                cov_title = f"covered-{title}"
                cov_val = covered_bdown[title]
                values[cov_title] = cov_val
                report_run_titles.append(cov_title)
                if title.endswith("-gbp"):
                    if isinstance(virt_val, (int, float, Decimal)):
                        if isinstance(cov_val, (int, float, Decimal)):
                            diff_val = float(cov_val) - float(virt_val)
                        else:
                            diff_val = 0 - float(virt_val)
                    else:
                        diff_val = 0

                    values[f"difference-{title}"] = diff_val

                    t = "difference-tpr-gbp"
                    try:
                        values[t] += diff_val
                    except KeyError:
                        values[t] = diff_val
                        report_run_titles.append(t)

        csv_row = []
        for t in titles:
            v = values[t]
            if t == "covered-bills":
                val = " | ".join(str(b) for b in v)
            else:
                val = csv_make_val(v)

            csv_row.append(val)

        for t in report_run_titles:
            if t not in titles:
                csv_row.append(t)
                csv_row.append(csv_make_val(values[t]))

        writer.writerow(csv_row)

        values["bill_id"] = bill.id
        values["batch_id"] = bill.batch.id
        values["supply_id"] = supply.id
        values["site_id"] = None if site_code is None else site.id
        ReportRun.w_insert_row(
            report_run_id, "", report_run_titles, values, {"is_checked": False}
        )

        for bill in sess.query(Bill).filter(
            Bill.supply == supply,
            Bill.start_date <= covered_finish,
            Bill.finish_date >= covered_start,
        ):
            for k, v in loads(bill.breakdown).items():
                if k.endswith("-gbp"):
                    add_gap(
                        caches,
                        gaps,
                        k[:-4],
                        bill.start_date,
                        bill.finish_date,
                        False,
                        v,
                    )

        # Avoid long-running transactions
        sess.rollback()

    clumps = []
    for element, elgap in sorted(gaps.items()):
        for start_date, hhgap in sorted(elgap.items()):
            if hhgap["has_virtual"] and not hhgap["has_covered"]:
                if len(clumps) == 0 or not all(
                    (
                        clumps[-1]["element"] == element,
                        clumps[-1]["finish_date"] + HH == start_date,
                    )
                ):
                    clumps.append(
                        {
                            "element": element,
                            "start_date": start_date,
                            "finish_date": start_date,
                            "gbp": hhgap["gbp"],
                        }
                    )
                else:
                    clumps[-1]["finish_date"] = start_date

    for i, clump in enumerate(clumps):
        vals = {}
        for title in titles:
            if title.startswith("difference-") and title.endswith("-gbp"):
                vals[title] = 0
            else:
                vals[title] = None

        vals["covered-problem"] = "_".join(
            (
                "missing",
                clump["element"],
                "supplyid",
                str(supply.id),
                "from",
                hh_format(clump["start_date"]),
            )
        )
        vals["imp-mpan-core"] = imp_mpan_core
        vals["exp-mpan-core"] = exp_mpan_core
        vals["batch"] = "missing_bill"
        vals["bill-start-date"] = hh_format(clump["start_date"])
        vals["bill-finish-date"] = hh_format(clump["finish_date"])
        vals["difference-net-gbp"] = clump["gbp"]
        writer.writerow(csv_make_val(vals[title]) for title in titles)

        vals["bill_id"] = None
        vals["batch_id"] = None
        vals["supply_id"] = supply.id
        vals["site_id"] = None if site_code is None else site.id

        ReportRun.w_insert_row(report_run_id, "", titles, vals, {"is_checked": False})

    # Avoid long-running transactions
    sess.rollback()



import csv
import sys
import threading
import traceback
from io import StringIO

from flask import g, request

from sqlalchemy import select
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.models import Contract, Era, ReportRun, Session, Site, SiteEra, User
from chellow.views import chellow_redirect

STATUSES_ACTIVE = ("IN USE / IN SERVICE", "STORED SPARE")
STATUSES_INACTIVE = ("DEMOLISHED", "SOLD", "ABANDONED")
STATUSES_IGNORE = (
    "OUT OF SERVICE",
    "SITE BEING CHECKED",
    "UNKNOWN",
    "UNADOPTED",
    "UNDER CONSTRUCTION",
    "EMERGENCY",
    "",
)


def _process_sites(sess, file_like, writer, props, report_run):
    ASSET_KEY = "asset_comparison"
    try:
        asset_props = props[ASSET_KEY]
    except KeyError:
        raise BadRequest(
            f"The property {ASSET_KEY} cannot be found in the configuration properties."
        )

    if not isinstance(asset_props, dict):
        raise BadRequest("The {ASSET_KEY} property must be a map.")

    for key in ("ignore_site_codes",):
        try:
            asset_props[key]
        except KeyError:
            raise BadRequest(
                f"The property {key} cannot be found in the '{ASSET_KEY}' section of "
                f"the configuration properties."
            )

    ignore_site_codes = asset_props["ignore_site_codes"]

    site_codes_select = (
        select(Site.code)
        .filter(Site.code.notin_(ignore_site_codes))
        .order_by(Site.code)
    )
    site_codes = [s[0] for s in sess.execute(site_codes_select)]

    titles = (
        "site_code",
        "asset_status",
        "chellow_status",
    )
    writer.writerow(titles)

    parser = iter(csv.reader(file_like))
    next(parser)  # Skip titles

    for values in parser:
        if len(values) == 0:
            continue

        asset_code = values[0].strip()
        asset_status = values[3].strip()

        if asset_code in ignore_site_codes:
            continue

        if asset_code in site_codes:
            site_codes.remove(asset_code)

        eras = sess.execute(
            select(Era)
            .join(SiteEra)
            .join(Site)
            .filter(Site.code == asset_code, Era.finish_date == null())
            .options(joinedload(Era.energisation_status))
        ).all()

        if len(eras) == 0:
            is_energised = None
            is_energised_str = "no supplies"
        else:
            energised_eras = [r for r in eras if r[0].energisation_status.code == "E"]
            is_energised = len(energised_eras) > 0
            is_energised_str = "energised" if is_energised is True else "de-energised"

        if asset_status in STATUSES_ACTIVE:
            active_asset = True
        elif asset_status in STATUSES_INACTIVE:
            active_asset = False
        elif asset_status in STATUSES_IGNORE:
            active_asset = None
        else:
            raise BadRequest(f"Asset status '{asset_status}' not recognized.")

        if active_asset is False and is_energised is not None:
            values = {
                "site_code": asset_code,
                "asset_status": asset_status,
                "chellow_status": is_energised_str,
            }
            writer.writerow([values[t] for t in titles])
            site = Site.find_by_code(sess, asset_code)
            values["site_id"] = None if site is None else site.id
            report_run.insert_row(sess, "", titles, values, {})
            sess.commit()

    for site_code in site_codes:
        eras = sess.execute(
            select(Era)
            .join(SiteEra)
            .join(Site)
            .filter(Site.code == site_code, Era.finish_date == null())
            .options(joinedload(Era.energisation_status))
        ).all()

        if len(eras) > 0:
            e_eras = [r for r in eras if r[0].energisation_status.code == "E"]
            is_energised_str = "energised" if len(e_eras) > 0 else "de-energised"
            values = {
                "site_code": site_code,
                "asset_status": None,
                "chellow_status": is_energised_str,
            }
            writer.writerow([values[t] for t in titles])
            site = Site.find_by_code(sess, site_code)
            values["site_id"] = None if site is None else site.id
            report_run.insert_row(sess, "", titles, values, {})
            sess.commit()


FNAME = "asset_comparison"


def content(user_id, file_like, report_run_id):
    f = report_run = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file(f"{FNAME}.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            report_run = ReportRun.get_by_id(sess, report_run_id)

            props = Contract.get_non_core_by_name(
                sess, "configuration"
            ).make_properties()

            _process_sites(sess, file_like, writer, props, report_run)
            report_run.update("finished")
            sess.commit()
    except BaseException:
        msg = traceback.format_exc()
        if report_run is not None:
            report_run.update("interrupted")
            report_run.insert_row(sess, "", ["error"], {"error": msg}, {})
            sess.commit()
        sys.stderr.write(msg)
        if writer is not None:
            writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_post(sess):
    user = g.user
    file_item = request.files["asset_file"]

    report_run = ReportRun.insert(
        sess,
        FNAME,
        user,
        FNAME,
        {
            "STATUSES_ACTIVE": STATUSES_ACTIVE,
            "STATUSES_INACTIVE": STATUSES_INACTIVE,
            "STATUSES_IGNORE": STATUSES_IGNORE,
        },
    )
    sess.commit()
    args = user.id, StringIO(file_item.read().decode("utf8")), report_run.id
    threading.Thread(target=content, args=args).start()
    return chellow_redirect(f"/report_runs/{report_run.id}", 303)



import sys
import threading
import traceback

from flask import g, request

import odio

from sqlalchemy import or_, select, true
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

import chellow.e.computer
from chellow.dloads import open_file
from chellow.e.computer import contract_func
from chellow.gas.engine import GDataSource
from chellow.models import (
    GBill,
    GContract,
    GEra,
    GSupply,
    Session,
    Site,
    SiteGEra,
)
from chellow.utils import (
    c_months_u,
    ct_datetime_now,
    hh_format,
    hh_max,
    hh_min,
    make_val,
    req_bool,
    req_int,
)
from chellow.views import chellow_redirect


CATEGORY_ORDER = {None: 0, "unmetered": 1, "nhh": 2, "amr": 3, "hh": 4}
meter_order = {"hh": 0, "amr": 1, "nhh": 2, "unmetered": 3}


def write_spreadsheet(fl, compressed, site_rows, era_rows):
    fl.seek(0)
    fl.truncate()
    with odio.create_spreadsheet(fl, "1.2", compressed=compressed) as f:
        f.append_table("Site Level", site_rows)
        f.append_table("Era Level", era_rows)


def _process_era(
    report_context,
    sess,
    site,
    g_era,
    month_start,
    month_finish,
    forecast_from,
    g_era_rows,
    vb_titles,
    now,
):
    g_supply = g_era.g_supply

    ss_start = hh_max(g_era.start_date, month_start)
    ss_finish = hh_min(g_era.finish_date, month_finish)

    ss = GDataSource(
        sess,
        ss_start,
        ss_finish,
        forecast_from,
        g_era,
        report_context,
        None,
    )

    contract = g_era.g_contract
    vb_function = contract_func(report_context, contract, "virtual_bill")
    if vb_function is None:
        raise BadRequest(
            f"The contract {contract.name} doesn't have the virtual_bill() function."
        )
    try:
        vb_function(ss)
        bill = ss.bill
    except TypeError as e:
        raise BadRequest(
            f"Problem with virtual bill for g_supplier_contrat {contract.id}."
        ) from e

    try:
        gbp = bill["net_gbp"]
    except KeyError:
        gbp = 0
        bill["problem"] += (
            f"For the supply {ss.mprn} the virtual bill {bill} "
            f"from the contract {contract.name} does not contain "
            f"the net_gbp key."
        )
    try:
        kwh = bill["kwh"]
    except KeyError:
        kwh = 0
        bill["problem"] += (
            f"For the supply {ss.mprn} the virtual bill "
            f"{bill} from the contract {contract.name} does not "
            f"contain the 'kwh' key."
        )

    billed_kwh = billed_net_gbp = billed_vat_gbp = billed_gross_gbp = 0

    g_era_associates = {s.site.code for s in g_era.site_g_eras if not s.is_physical}

    for g_bill in sess.query(GBill).filter(
        GBill.g_supply == g_supply,
        GBill.start_date <= ss_finish,
        GBill.finish_date >= ss_start,
    ):
        bill_start = g_bill.start_date
        bill_finish = g_bill.finish_date
        bill_duration = (bill_finish - bill_start).total_seconds() + (30 * 60)
        overlap_duration = (
            min(bill_finish, ss_finish) - max(bill_start, ss_start)
        ).total_seconds() + (30 * 60)
        overlap_proportion = overlap_duration / bill_duration
        billed_kwh += overlap_proportion * float(g_bill.kwh)
        billed_net_gbp += overlap_proportion * float(g_bill.net)
        billed_vat_gbp += overlap_proportion * float(g_bill.vat)
        billed_gross_gbp += overlap_proportion * float(g_bill.gross)

    associated_site_ids = ",".join(sorted(g_era_associates))
    g_era_rows.append(
        [
            make_val(v)
            for v in [
                now,
                g_supply.mprn,
                g_supply.name,
                g_supply.g_exit_zone.code,
                g_era.msn,
                g_era.g_unit.code,
                contract.name,
                site.code,
                site.name,
                associated_site_ids,
                g_era.start_date,
                month_finish,
                kwh,
                gbp,
                billed_kwh,
                billed_net_gbp,
                billed_vat_gbp,
                billed_gross_gbp,
            ]
        ]
        + [make_val(bill.get(t)) for t in vb_titles]
    )
    return kwh, gbp, billed_kwh, billed_net_gbp, billed_vat_gbp, billed_gross_gbp


def content(
    site_id, g_supply_id, user, compression, finish_year, finish_month, months, now=None
):
    if now is None:
        now = ct_datetime_now()
    report_context = {}
    month_list = list(
        c_months_u(finish_year=finish_year, finish_month=finish_month, months=months)
    )
    start_date, finish_date = month_list[0][0], month_list[-1][-1]

    try:
        with Session() as sess:
            base_name = [
                "g_monthly_duration",
                hh_format(start_date)
                .replace(" ", "_")
                .replace(":", "")
                .replace("-", ""),
                "for",
                str(months),
                "months",
            ]

            forecast_from = chellow.e.computer.forecast_date()

            sites = (
                sess.query(Site)
                .join(SiteGEra)
                .join(GEra)
                .filter(SiteGEra.is_physical == true())
                .distinct()
                .order_by(Site.code)
            )
            if site_id is not None:
                site = Site.get_by_id(sess, site_id)
                sites = sites.filter(Site.id == site.id)
                base_name.append("site")
                base_name.append(site.code)
            if g_supply_id is not None:
                g_supply = GSupply.get_by_id(sess, g_supply_id)
                base_name.append("g_supply")
                base_name.append(str(g_supply.id))
                sites = sites.filter(GEra.g_supply == g_supply)

            rf = open_file("_".join(base_name) + ".ods", user, mode="wb")
            site_rows = []
            g_era_rows = []

            era_header_titles = [
                "creation_date",
                "mprn",
                "supply_name",
                "exit_zone",
                "msn",
                "unit",
                "contract",
                "site_id",
                "site_name",
                "associated_site_ids",
                "era-start",
                "month",
            ]
            site_header_titles = [
                "creation_date",
                "site_id",
                "site_name",
                "associated_site_ids",
                "month",
            ]
            summary_titles = [
                "kwh",
                "gbp",
                "billed_kwh",
                "billed_net_gbp",
                "billed_vat_gbp",
                "billed_gross_gbp",
            ]

            vb_titles = []
            conts = (
                sess.query(GContract)
                .join(GEra)
                .join(GSupply)
                .filter(
                    GEra.start_date <= finish_date,
                    or_(GEra.finish_date == null(), GEra.finish_date >= start_date),
                )
                .distinct()
                .order_by(GContract.id)
            )
            if g_supply_id is not None:
                conts = conts.filter(GEra.g_supply_id == g_supply_id)
            for cont in conts:
                title_func = chellow.e.computer.contract_func(
                    report_context, cont, "virtual_bill_titles"
                )
                if title_func is None:
                    raise Exception(
                        f"For the contract {cont.name} there doesn't seem to be "
                        f"a 'virtual_bill_titles' function."
                    )
                for title in title_func():
                    if title not in vb_titles:
                        vb_titles.append(title)

            g_era_rows.append(era_header_titles + summary_titles + vb_titles)
            site_rows.append(site_header_titles + summary_titles)

            for month_start, month_finish in month_list:
                for site in sites.filter(
                    GEra.start_date <= month_finish,
                    or_(GEra.finish_date == null(), GEra.finish_date >= month_start),
                ):
                    site_kwh = site_gbp = site_billed_kwh = site_billed_net_gbp = 0
                    site_billed_vat_gbp = site_billed_gross_gbp = 0

                    g_eras_q = (
                        select(GEra)
                        .join(SiteGEra)
                        .filter(
                            SiteGEra.site == site,
                            SiteGEra.is_physical == true(),
                            GEra.start_date <= month_finish,
                            or_(
                                GEra.finish_date == null(),
                                GEra.finish_date >= month_start,
                            ),
                        )
                        .options(
                            joinedload(GEra.g_contract),
                            joinedload(GEra.g_supply),
                            joinedload(GEra.g_supply).joinedload(GSupply.g_exit_zone),
                        )
                        .order_by(GEra.id)
                    )
                    if g_supply_id is not None:
                        g_eras_q = g_eras_q.where(GEra.g_supply_id == g_supply_id)

                    for g_era in sess.scalars(g_eras_q):
                        try:
                            (
                                kwh,
                                gbp,
                                billed_kwh,
                                billed_net_gbp,
                                billed_vat_gbp,
                                billed_gross_gbp,
                            ) = _process_era(
                                report_context,
                                sess,
                                site,
                                g_era,
                                month_start,
                                month_finish,
                                forecast_from,
                                g_era_rows,
                                vb_titles,
                                now,
                            )
                            site_kwh += kwh
                            site_gbp += gbp
                            site_billed_kwh += billed_kwh
                            site_billed_net_gbp += billed_net_gbp
                            site_billed_vat_gbp += billed_vat_gbp
                            site_billed_gross_gbp += billed_gross_gbp
                        except BadRequest as e:
                            raise BadRequest(
                                f"Problem with g_era {g_era.id}: {e.description}"
                            )

                    linked_sites = ", ".join(
                        s.code
                        for s in site.find_linked_sites(sess, month_start, month_finish)
                    )

                    site_rows.append(
                        [
                            make_val(v)
                            for v in [
                                now,
                                site.code,
                                site.name,
                                linked_sites,
                                month_finish,
                                site_kwh,
                                site_gbp,
                                site_billed_kwh,
                                site_billed_net_gbp,
                                site_billed_vat_gbp,
                                site_billed_gross_gbp,
                            ]
                        ]
                    )
                    sess.rollback()
                write_spreadsheet(rf, compression, site_rows, g_era_rows)

    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + "\n")
        site_rows.append(["Problem " + msg])
        write_spreadsheet(rf, compression, site_rows, g_era_rows)
    finally:
        try:
            rf.close()
        except BaseException:
            msg = traceback.format_exc()
            ef = open_file("error.txt", user, mode="w")
            ef.write(msg + "\n")
            ef.close()


def do_get(sess):
    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")
    months = req_int("months")

    site_id = req_int("site_id") if "site_id" in request.values else None

    if "g_supply_id" in request.values:
        g_supply_id = req_int("g_supply_id")
    else:
        g_supply_id = None

    if "compression" in request.values:
        compression = req_bool("compression")
    else:
        compression = True

    user = g.user
    args = (site_id, g_supply_id, user, compression, finish_year, finish_month, months)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from decimal import Decimal

import xlrd.sheet

from chellow.e.bill_parsers.sww_xls import _parse_row
from chellow.utils import utc_datetime


def test_parse_row(mocker):
    row = []
    for val in [
        "2019-04-01",
        "2019-04-30",
        "22 7673 2894 214",
        54,
        7.8,
        74,
        5998,
        6.22,
        52,
        877,
        8.001,
        82.98,
        377,
        0.3,
        52,
        823,
        0.782,
        305,
        672,
        4.62,
        8483,
        6843,
        6.328,
        6984,
        7034,
        6734,
        574,
        956,
        678,
        4567,
        485,
        57,
        76,
        39647,
        6587,
        768,
        34687,
        67,
        5478,
        879654,
        654897,
        54789,
        None,
        None,
        None,
        43895,
        48,
        498,
        3498,
        9384,
        23,
    ]:
        cell = mocker.Mock(spec=xlrd.sheet.Cell)
        cell.value = val
        row.append(cell)

    row_index = 2
    datemode = 0
    titles = [
        "BILL START DATE",
        "BILL END DATE",
        "MPAN",
        "DUOS-AVAIL-AV USE KVADAY",
        "DUOS-AVAIL-AV RATE P/KVADAY",
        "DUOS-AVAIL-AV COST GBP",
        "DUOS-STND-SITE USE DAY",
        "DUOS-STND-SITE RATE P/DAY",
        "DUOS-STND-SITE COST GBP",
        "DUOS-UNIT-AMBER USE KWH",
        "DUOS-UNIT-AMBER RATE P/KWH",
        "DUOS-UNIT-AMBER COST GBP",
        "DUOS-UNIT-GREEN USE KWH",
        "DUOS-UNIT-GREEN RATE P/KWH",
        "DUOS-UNIT-GREEN COST GBP",
        "DUOS-UNIT-RED USE KWH",
        "DUOS-UNIT-RED RATE P/KWH",
        "DUOS-UNIT-RED COST GBP",
        "LVY-AAHEDC-ALL USE KWH",
        "LVY-AAHEDC-ALL RATE P/KWH",
        "LVY-AAHEDC-ALL COST GBP",
        "LVY-BSUOS-ALL USE KWH",
        "LVY-BSUOS-ALL RATE P/KWH",
        "LVY-BSUOS-ALL COST GBP",
        "LVY-CCL-ALL USE KWH",
        "LVY-CCL-ALL RATE P/KWH",
        "LVY-CCL-ALL COST GBP",
        "LVY-CFDOB-ALL USE KWH",
        "LVY-CFDOB-ALL RATE P/KWH",
        "LVY-CFDOB-ALL COST GBP",
        "LVY-CFDOP-ALL USE KWH",
        "LVY-CFDOP-ALL RATE P/KWH",
        "LVY-CFDOP-ALL COST GBP",
        "LVY-FIT-ALL USE KWH",
        "LVY-FIT-ALL RATE P/KWH",
        "LVY-FIT-ALL COST GBP",
        "LVY-RO-ALL USE KWH",
        "LVY-RO-ALL RATE P/KWH",
        "LVY-RO-ALL COST GBP",
        "NRG-UNIT-SUMMER USE KWH",
        "NRG-UNIT-SUMMER RATE P/KWH",
        "NRG-UNIT-SUMMER COST GBP",
        "NRG-UNIT-WINTER USE KWH",
        "NRG-UNIT-WINTER RATE P/KWH",
        "NRG-UNIT-WINTER COST GBP",
        "SYS-ADMIN-ALL USE MO",
        "SYS-ADMIN-ALL RATE GBP/MO",
        "SYS-ADMIN-ALL COST GBP",
        "TUOS-TRIAD-SITE USE DAY",
        "TUOS-TRIAD-SITE RATE GBP/DAY",
        "TUOS-TRIAD-SITE COST GBP",
    ]
    title_row = []
    for title in titles:
        cell = mocker.Mock(spec=xlrd.sheet.Cell)
        cell.value = title
        title_row.append(cell)

    bill = _parse_row(row, row_index, datemode, title_row)

    bill_start_date = utc_datetime(2019, 4, 1)
    assert bill["start_date"] == bill_start_date

    bill_finish_date = utc_datetime(2019, 4, 30, 23, 30)
    assert bill["finish_date"] == bill_finish_date

    assert bill["net"] == Decimal("82805.98")
    assert bill["breakdown"]["aahedc-gbp"] == Decimal("8483")
    assert bill["mpan_core"] == "22 0003 0354 632"



from chellow.e.computer import (
    SiteSource,
    SupplySource,
    _find_hhs,
    _find_pair,
    _init_hh_data,
    _make_reads,
    _set_status,
)
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcLlfcSsc,
    MtcLlfcSscPc,
    MtcParticipant,
    MtcSsc,
    Participant,
    Pc,
    Site,
    Source,
    Ssc,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_SiteSource_init_hh_data(sess, mocker):
    """New style channels"""
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        vf,
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(sess, "Fus DC", participant, "", {}, vf, None, {})
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fus Sup", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(sess, "510", "HH HV", voltage_level, False, True, vf, None)
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    channel = era.insert_channel(sess, True, "ACTIVE")
    data_raw = [
        {
            "start_date": utc_datetime(2009, 8, 10),
            "value": 10,
            "status": "A",
        }
    ]
    channel.add_hh_data(sess, data_raw)

    sess.commit()

    caches = {}
    start_date = utc_datetime(2009, 7, 31, 23, 00)
    finish_date = utc_datetime(2009, 7, 31, 23, 00)
    fdate = utc_datetime(2010, 7, 31, 23, 00)
    ds = SiteSource(sess, site, start_date, finish_date, fdate, caches)

    actual_datum = ds.hh_data[0]
    expected_datum = {
        "hist-start": utc_datetime(2009, 7, 31, 23, 0),
        "start-date": utc_datetime(2009, 7, 31, 23, 0),
        "ct-day": 1,
        "utc-month": 7,
        "utc-day": 31,
        "utc-decimal-hour": 23.0,
        "utc-year": 2009,
        "utc-hour": 23,
        "utc-minute": 0,
        "ct-year": 2009,
        "ct-month": 8,
        "ct-decimal-hour": 0.0,
        "ct-day-of-week": 5,
        "utc-day-of-week": 4,
        "utc-is-bank-holiday": False,
        "ct-is-bank-holiday": False,
        "utc-is-month-end": False,
        "ct-is-month-end": False,
        "status": "E",
        "imp-msp-kvarh": 0,
        "imp-msp-kvar": 0,
        "exp-msp-kvarh": 0,
        "exp-msp-kvar": 0,
        "msp-kva": 0,
        "msp-kw": 0,
        "msp-kwh": 0,
        "hist-import-grid-kvarh": 0,
        "hist-export-grid-kvarh": 0,
        "anti-msp-kwh": 0,
        "anti-msp-kw": 0,
        "hist-imp-msp-kvarh": 0,
        "hist-kwh": 0,
        "hist-import-grid-kwh": 0,
        "hist-export-grid-kwh": 0,
        "hist-import-gen-kwh": 0,
        "hist-export-gen-kwh": 0,
        "hist-import-3rd-party-kwh": 0,
        "hist-export-3rd-party-kwh": 0,
        "hist-used-3rd-party-kwh": 0,
        "used-3rd-party-kwh": 0,
        "hist-used-gen-msp-kwh": 0,
        "used-gen-msp-kwh": 0,
        "import-grid-kwh": 0,
        "export-grid-kwh": 0,
        "import-gen-kwh": 0,
        "export-gen-kwh": 0,
        "import-3rd-party-kwh": 0,
        "export-3rd-party-kwh": 0,
        "hist-used-kwh": 0,
        "used-kwh": 0,
        "used-gen-msp-kw": 0,
    }
    assert actual_datum == expected_datum


def test_find_pair(mocker):
    sess = mocker.Mock()
    caches = {}
    is_forwards = True
    first_read = {
        "date": utc_datetime(2010, 1, 1),
        "reads": {},
        "msn": "kh",
        "read_type": "N",
    }
    second_read = {
        "date": utc_datetime(2010, 2, 1),
        "reads": {},
        "msn": "kh",
        "read_type": "N",
    }
    read_list = [first_read, second_read]
    pair = _find_pair(sess, caches, is_forwards, read_list)
    assert pair["start-date"] == utc_datetime(2010, 1, 1)


def test_find_hhs_empty_pairs(mocker):
    mocker.patch("chellow.e.computer.is_tpr", return_value=True)
    caches = {}
    sess = mocker.Mock()
    pairs = []
    chunk_start = utc_datetime(2010, 1, 1)
    chunk_finish = utc_datetime(2010, 1, 1)
    hhs = _find_hhs(caches, sess, pairs, chunk_start, chunk_finish)
    assert hhs == {
        utc_datetime(2010, 1, 1): {
            "msp-kw": 0,
            "msp-kwh": 0,
            "hist-kwh": 0,
            "imp-msp-kvar": 0,
            "imp-msp-kvarh": 0,
            "exp-msp-kvar": 0,
            "exp-msp-kvarh": 0,
            "tpr": "00001",
        }
    }


def test_find_hhs_two_pairs(mocker):
    mocker.patch("chellow.e.computer.is_tpr", return_value=True)
    caches = {}
    sess = mocker.Mock()
    pairs = [
        {"start-date": utc_datetime(2010, 1, 1), "tprs": {"00001": 1}},
        {"start-date": utc_datetime(2010, 1, 1, 0, 30), "tprs": {"00001": 1}},
    ]
    chunk_start = utc_datetime(2010, 1, 1)
    chunk_finish = utc_datetime(2010, 1, 1, 0, 30)
    hhs = _find_hhs(caches, sess, pairs, chunk_start, chunk_finish)
    assert hhs == {
        utc_datetime(2010, 1, 1): {
            "msp-kw": 2.0,
            "msp-kwh": 1.0,
            "hist-kwh": 1.0,
            "imp-msp-kvar": 0,
            "imp-msp-kvarh": 0,
            "exp-msp-kvar": 0,
            "exp-msp-kvarh": 0,
            "tpr": "00001",
        },
        utc_datetime(2010, 1, 1, 0, 30): {
            "msp-kw": 2.0,
            "msp-kwh": 1.0,
            "hist-kwh": 1.0,
            "imp-msp-kvar": 0,
            "imp-msp-kvarh": 0,
            "exp-msp-kvar": 0,
            "exp-msp-kvarh": 0,
            "tpr": "00001",
        },
    }


def test_set_status(mocker):
    hhs = {utc_datetime(2012, 2, 1): {}}

    read_list = [{"date": utc_datetime(2012, 1, 1)}]
    forecast_date = utc_datetime(2012, 3, 1)
    _set_status(hhs, read_list, forecast_date)
    assert hhs == {utc_datetime(2012, 2, 1): {"status": "E"}}


def test_make_reads_forwards(mocker):
    is_forwards = True
    msn = "k"
    read_a = {"date": utc_datetime(2018, 3, 10), "msn": msn}
    read_b = {"date": utc_datetime(2018, 3, 13), "msn": msn}
    prev_reads = iter([read_a])
    pres_reads = iter([read_b])
    actual = list(_make_reads(is_forwards, prev_reads, pres_reads))
    expected = [read_a, read_b]
    assert actual == expected


def test_make_reads_forwards_meter_change(mocker):
    is_forwards = True
    dt = utc_datetime(2018, 3, 1)
    read_a = {"date": dt, "msn": "a"}
    read_b = {"date": dt, "msn": "b"}
    prev_reads = iter([read_a])
    pres_reads = iter([read_b])
    actual = list(_make_reads(is_forwards, prev_reads, pres_reads))
    expected = [read_b, read_a]
    assert actual == expected


def test_make_reads_backwards(mocker):
    is_forwards = False
    msn = "k"
    read_a = {"date": utc_datetime(2018, 3, 10), "msn": msn}
    read_b = {"date": utc_datetime(2018, 3, 13), "msn": msn}
    prev_reads = iter([read_a])
    pres_reads = iter([read_b])
    actual = list(_make_reads(is_forwards, prev_reads, pres_reads))
    expected = [read_b, read_a]
    assert actual == expected


def test_init_hh_data(sess, mocker):
    """New style channels"""
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        vf,
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    channel = era.insert_channel(sess, True, "ACTIVE")
    data_raw = [
        {
            "start_date": utc_datetime(2009, 8, 10),
            "value": 10,
            "status": "A",
        }
    ]
    channel.add_hh_data(sess, data_raw)

    sess.commit()

    caches = {}
    chunk_start = utc_datetime(2009, 7, 31, 23, 00)
    chunk_finish = utc_datetime(2009, 8, 31, 22, 30)
    is_import = True
    full_channels, hhd = _init_hh_data(
        sess, caches, era, chunk_start, chunk_finish, is_import
    )

    assert full_channels

    expected_hhd = {
        utc_datetime(2009, 8, 10): {
            "imp-msp-kvarh": 0.0,
            "imp-msp-kvar": 0.0,
            "exp-msp-kvarh": 0.0,
            "exp-msp-kvar": 0.0,
            "status": "A",
            "hist-kwh": 10.0,
            "msp-kwh": 10.0,
            "msp-kw": 20.0,
        }
    }
    assert hhd == expected_hhd


def test_init_hh_data_export(sess, mocker):
    """New style channels"""
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    exp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "521",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        False,
        valid_from,
        None,
    )
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        None,
        None,
        None,
        None,
        None,
        "22 7867 6232 781",
        "521",
        exp_supplier_contract,
        "7748",
        361,
    )
    era = supply.eras[0]
    active_channel = era.insert_channel(sess, False, "ACTIVE")
    active_data_raw = [
        {
            "start_date": utc_datetime(2009, 8, 10),
            "value": 10,
            "status": "A",
        }
    ]
    active_channel.add_hh_data(sess, active_data_raw)

    reactive_channel = era.insert_channel(sess, False, "REACTIVE_IMP")
    reactive_data_raw = [
        {
            "start_date": utc_datetime(2009, 8, 10),
            "value": 5,
            "status": "A",
        }
    ]
    reactive_channel.add_hh_data(sess, reactive_data_raw)
    sess.commit()

    caches = {}
    chunk_start = utc_datetime(2009, 7, 31, 23, 00)
    chunk_finish = utc_datetime(2009, 8, 31, 22, 30)
    is_import = False
    full_channels, hhd = _init_hh_data(
        sess, caches, era, chunk_start, chunk_finish, is_import
    )

    assert full_channels

    expected_hhd = {
        utc_datetime(2009, 8, 10): {
            "imp-msp-kvarh": 5.0,
            "imp-msp-kvar": 10.0,
            "exp-msp-kvarh": 0.0,
            "exp-msp-kvar": 0.0,
            "status": "A",
            "hist-kwh": 10.0,
            "msp-kwh": 10.0,
            "msp-kw": 20.0,
        }
    }
    assert hhd == expected_hhd


def test_SupplySource_init_hh(sess, mocker):
    """Old style channels"""
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    exp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "521",
        "Export (HV)",
        voltage_level,
        False,
        False,
        utc_datetime(1996, 1, 1),
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        None,
        None,
        None,
        None,
        None,
        "22 7867 6232 781",
        "521",
        exp_supplier_contract,
        "7748",
        361,
    )
    era = supply.eras[0]
    active_channel = era.insert_channel(sess, False, "ACTIVE")
    active_data_raw = [
        {
            "start_date": utc_datetime(2009, 8, 10),
            "value": 10,
            "status": "A",
        }
    ]
    active_channel.add_hh_data(sess, active_data_raw)
    era.insert_channel(sess, True, "REACTIVE_IMP")

    sess.commit()

    caches = {}
    start_date = utc_datetime(2009, 7, 31, 23, 00)
    finish_date = utc_datetime(2009, 8, 31, 22, 30)
    forecast_date = utc_datetime(2019, 8, 31, 22, 30)
    is_import = False
    ss = SupplySource(
        sess, start_date, finish_date, forecast_date, era, is_import, caches
    )

    assert not ss.full_channels


def test_SupplySource_init_nhh(sess, mocker):
    """Old style channels"""
    valid_from = utc_datetime(1996, 1, 1)
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_X, "Fusion Ltc", utc_datetime(2000, 1, 1), None, None
    )
    participant.insert_party(
        sess, market_role_C, "Fusion DC", utc_datetime(2000, 1, 1), None, None
    )
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, utc_datetime(2000, 1, 1), None, {}
    )
    pc = Pc.insert(sess, "03", "nhh", utc_datetime(2000, 1, 1), None)
    ssc = Ssc.insert(sess, "0393", "unrestricted", True, utc_datetime(2000, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    exp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(
        sess, market_role_R, "WPD", utc_datetime(2000, 1, 1), None, "22"
    )
    Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(
        sess, "CR", "Credit", utc_datetime(1996, 1, 1), None
    )
    mtc = Mtc.insert(
        sess,
        "845",
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc_imp = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        utc_datetime(1996, 1, 1),
        None,
    )
    llfc_exp = dno.insert_llfc(
        sess,
        "521",
        "Export (HV)",
        voltage_level,
        False,
        False,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "D")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, valid_from, None)
    MtcLlfc.insert(sess, mtc_participant, llfc_imp, valid_from, None)
    mtc_llfc_imp_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc_imp, valid_from, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_imp_ssc, pc, valid_from, None)
    MtcLlfc.insert(sess, mtc_participant, llfc_exp, valid_from, None)
    mtc_llfc_exp_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc_exp, valid_from, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_exp_ssc, pc, valid_from, None)
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        "773",
        dc_contract,
        "ghyy3",
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        ssc.code,
        energisation_status,
        dtc_meter_type,
        None,
        None,
        None,
        None,
        None,
        "22 7867 6232 781",
        "521",
        exp_supplier_contract,
        "7748",
        361,
    )
    era = supply.eras[0]
    active_channel = era.insert_channel(sess, False, "ACTIVE")
    active_data_raw = [
        {
            "start_date": utc_datetime(2009, 8, 10),
            "value": 10,
            "status": "A",
        }
    ]
    active_channel.add_hh_data(sess, active_data_raw)
    era.insert_channel(sess, True, "REACTIVE_IMP")

    sess.commit()

    caches = {}
    start_date = utc_datetime(2009, 7, 31, 23, 00)
    finish_date = utc_datetime(2009, 8, 31, 22, 30)
    forecast_date = utc_datetime(2019, 8, 31, 22, 30)
    is_import = False
    SupplySource(sess, start_date, finish_date, forecast_date, era, is_import, caches)


def test_SiteSource_get_data_sources(mocker):
    mocker.patch.object(SiteSource, "__init__", lambda *x: None)
    mocker.patch("chellow.e.computer.displaced_era")
    ds = SiteSource()
    ds.forecast_date = to_utc(ct_datetime(2010, 1, 1))
    ds.start_date = to_utc(ct_datetime(2008, 1, 1))
    ds.finish_date = to_utc(ct_datetime(2008, 8, 31, 22, 30))
    ds.stream_focus = "gen-used"
    ds.sess = mocker.Mock()
    ds.caches = {}
    ds.site = mocker.Mock()
    ds.era_maps = mocker.Mock()
    ds.deltas = mocker.Mock()

    start_date = to_utc(ct_datetime(2008, 7, 1))
    finish_date = to_utc(ct_datetime(2008, 7, 31, 23, 30))
    result = ds.get_data_sources(start_date, finish_date)
    next(result)


def test_SiteSource_get_data_sources_clock_change(mocker):
    mocker.patch.object(SiteSource, "__init__", lambda *x: None)
    mocker.patch("chellow.e.computer.displaced_era")
    mock_c_months_u = mocker.patch("chellow.e.computer.c_months_u")
    ds = SiteSource()
    ds.forecast_date = to_utc(ct_datetime(2010, 1, 1))
    ds.start_date = to_utc(ct_datetime(2008, 1, 1))
    ds.finish_date = to_utc(ct_datetime(2008, 8, 31, 22, 30))
    ds.stream_focus = "gen-used"
    ds.sess = mocker.Mock()
    ds.caches = {}
    ds.site = mocker.Mock()
    ds.era_maps = mocker.Mock()
    ds.deltas = mocker.Mock()

    start_date = to_utc(ct_datetime(2008, 4, 1))
    finish_date = to_utc(ct_datetime(2008, 4, 30, 23, 30))
    result = ds.get_data_sources(start_date, finish_date)
    try:
        next(result)
    except StopIteration:
        pass
    mock_c_months_u.assert_called_with(
        finish_month=4, finish_year=2008, start_month=4, start_year=2008
    )



import atexit
import collections
import os
import os.path
import threading
import time
import traceback

from zipfile import ZIP_DEFLATED, ZipFile

from werkzeug.exceptions import BadRequest

from chellow.utils import utc_datetime_now


download_id = 0
lock = threading.Lock()

download_path = None
SERIAL_DIGITS = 5


def startup(instance_path, run_deleter=True):
    global file_deleter
    global download_id
    global download_path

    download_path = instance_path / "downloads"
    download_path.mkdir(parents=True, exist_ok=True)

    # Delete any files marked RUNNING that are hanging around
    for fl in download_path.iterdir():
        if fl.name[SERIAL_DIGITS:].startswith("_RUNNING_"):
            new_name = fl.name.replace("_RUNNING_", "_INTERRUPTED_", 1)
            fl.rename(fl.parent / new_name)

    dirs = sorted(download_path.iterdir(), reverse=True)
    if len(dirs) > 0:
        download_id = int(dirs[0].name[:SERIAL_DIGITS]) + 1

    if run_deleter:
        file_deleter = FileDeleter()
        file_deleter.start()


class DloadFile:
    def __init__(self, running_name, finished_name, mode, newline, is_zip):
        self.running_name = running_name
        self.finished_name = finished_name
        if is_zip:
            self.f = ZipFile(running_name, mode=mode, compression=ZIP_DEFLATED)
        else:
            self.f = self.running_name.open(mode=mode, newline=newline)

    def _check_exists(self):
        if not self.running_name.exists():
            raise BadRequest("Output file has been deleted.")

    def flush(self, *args, **kwargs):
        self._check_exists()
        return self.f.flush(*args, **kwargs)

    def write(self, *args, **kwargs):
        self._check_exists()
        return self.f.write(*args, **kwargs)

    def seek(self, *args, **kwargs):
        self._check_exists()
        return self.f.seek(*args, **kwargs)

    def truncate(self, *args, **kwargs):
        self._check_exists()
        return self.f.truncate(*args, **kwargs)

    def writestr(self, *args, **kwargs):
        self._check_exists()
        return self.f.writestr(*args, **kwargs)

    def close(self):
        self.f.close()
        self._check_exists()
        self.running_name.rename(self.finished_name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def open_file(base, user, mode="r", newline=None, is_zip=False):
    global download_id

    base = base.replace("/", "").replace(" ", "")
    try:
        lock.acquire()
        if len(list(download_path.iterdir())) == 0:
            download_id = 0
        serial = str(download_id).zfill(SERIAL_DIGITS)
        download_id += 1
    finally:
        lock.release()

    if user is None:
        uname = ""
    else:
        if hasattr(user, "proxy_username"):
            un = user.proxy_username
        else:
            un = user.email_address
        uname = un.replace("@", "").replace(".", "").replace("\\", "")

    names = tuple("_".join((serial, v, uname, base)) for v in ("RUNNING", "FINISHED"))
    running_name, finished_name = tuple(download_path / name for name in names)
    return DloadFile(running_name, finished_name, mode, newline, is_zip)


mem_id = 0
mem_lock = threading.Lock()
mem_vals = {}


def get_mem_id():
    global mem_id
    with mem_lock:
        mid = mem_id
        mem_id += 1
        mem_vals[mid] = None
    return mid


def put_mem_val(mem_id, val):
    with mem_lock:
        mem_vals[mem_id] = val


def get_mem_val(mem_id):
    with mem_lock:
        val = mem_vals.get(mem_id)
    return val


def get_mem_items():
    with mem_lock:
        return mem_vals.copy()


def remove_item(mem_id):
    with mem_lock:
        if mem_id in mem_vals:
            del mem_vals[mem_id]


file_deleter = None
MAX_AGE = 60 * 60 * 24 * 14


class FileDeleter(threading.Thread):
    def __init__(self):
        super(FileDeleter, self).__init__(name="File Deleter")
        self.lock = threading.RLock()
        self.messages = collections.deque()
        self.stopped = threading.Event()
        self.going = threading.Event()

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + " - " + message
        )
        if len(self.messages) > 100:
            self.messages.pop()

    def run(self):
        while not self.stopped.isSet():
            if self.lock.acquire(False):
                try:
                    cur_time = time.time()
                    for file_name in sorted(os.listdir(download_path)):
                        file_path = os.path.join(download_path, file_name)
                        if cur_time - os.path.getmtime(file_path) > MAX_AGE:
                            os.remove(file_path)
                except BaseException:
                    self.log("Outer problem " + traceback.format_exc())
                finally:
                    self.lock.release()
                    self.log("Finished deleting files.")

            self.going.wait(24 * 60 * 60)
            self.going.clear()


def get_file_deleter():
    return file_deleter


@atexit.register
def shutdown():
    if file_deleter is not None:
        file_deleter.stop()



import csv
import threading
import traceback
from collections import defaultdict

from flask import g

from sqlalchemy import select
from sqlalchemy.orm import joinedload

from chellow.dloads import open_file
from chellow.models import Batch, Bill, Session, User
from chellow.utils import csv_make_val, req_int
from chellow.views import chellow_redirect


def _content(sess, writer, batch_id):
    batch = Batch.get_by_id(sess, batch_id)
    titles = [
        "supplier_contract",
        "batch_reference",
        "bill_reference",
        "imp_mpan_core",
        "account",
        "issued",
        "from",
        "to",
        "kwh",
        "net",
        "vat",
        "gross",
        "type",
        "vat_1_percent",
        "vat_1_net",
        "vat_1_vat",
        "vat_2_percent",
        "vat_2_net",
        "vat_2_vat",
    ]
    writer.writerow(titles)

    for bill in sess.scalars(
        select(Bill)
        .where(Bill.batch == batch)
        .order_by(Bill.reference, Bill.start_date)
        .options(joinedload(Bill.bill_type), joinedload(Bill.supply))
    ):
        era = bill.supply.find_era_at(sess, bill.start_date)
        vat_breakdown = {}

        bd = bill.bd
        if "vat" in bd:
            for vat_percentage, vat_vals in bd["vat"].items():
                try:
                    vbd = vat_breakdown[vat_percentage]
                except KeyError:
                    vbd = vat_breakdown[vat_percentage] = defaultdict(int)

                vbd["vat"] += vat_vals["vat"]
                vbd["net"] += vat_vals["net"]

        vals = {
            "supplier_contract": batch.contract.name,
            "batch_reference": batch.reference,
            "bill_reference": bill.reference,
            "imp_mpan_core": None if era is None else era.imp_mpan_core,
            "account": bill.account,
            "issued": bill.issue_date,
            "from": bill.start_date,
            "to": bill.finish_date,
            "kwh": bill.kwh,
            "net": bill.net,
            "vat": bill.vat,
            "gross": bill.gross,
            "type": bill.bill_type.code,
            "vat_1_percent": None,
            "vat_1_net": None,
            "vat_1_vat": None,
            "vat_2_percent": None,
            "vat_2_net": None,
            "vat_2_vat": None,
        }
        for i, (percentage, vbd) in enumerate(sorted(vat_breakdown.items()), 1):
            vals[f"vat_{i}_percentage"] = percentage
            vals[f"vat_{i}_net"] = vbd["net"]
            vals[f"vat_{i}_vat"] = vbd["vat"]

        writer.writerow(csv_make_val(vals[t]) for t in titles)


def content(user_id, batch_id):
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file(f"bills_batch_{batch_id}.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            _content(sess, writer, batch_id)

    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        if f is not None:
            f.write(msg)
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    batch_id = req_int("batch_id")
    args = g.user.id, batch_id
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from io import StringIO

from chellow.models import Contract, MarketRole, Participant, User, UserRole

from chellow.reports.report_233 import content
from chellow.utils import ct_datetime, to_utc


def test_content(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id

    sess.commit()
    f = StringIO()
    mocker.patch("chellow.reports.report_233.open_file", return_value=f)
    days_hidden = 0
    is_ignored = True
    content(dc_contract.id, days_hidden, is_ignored, user_id)



import csv
import threading
import traceback

from flask import g

from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.e.computer import forecast_date
from chellow.gas.engine import GDataSource, g_contract_func
from chellow.models import GEra, GSupply, Session, Site, SiteGEra
from chellow.utils import csv_make_val, hh_format, hh_max, hh_min, req_date, req_int
from chellow.views import chellow_redirect


def content(g_supply_id, file_name, start_date, finish_date, user):
    caches = {}
    try:
        with Session() as sess:
            f = open_file(
                f"g_supply_virtual_bill_{g_supply_id}.csv", user, mode="w", newline=""
            )
            writer = csv.writer(f, lineterminator="\n")

            g_supply = GSupply.get_by_id(sess, g_supply_id)

            forecast_dt = forecast_date()

            prev_titles = None

            for g_era in (
                sess.query(GEra)
                .filter(
                    GEra.g_supply == g_supply,
                    GEra.start_date < finish_date,
                    or_(GEra.finish_date == null(), GEra.finish_date > start_date),
                )
                .order_by(GEra.start_date)
            ):
                chunk_start = hh_max(g_era.start_date, start_date)
                chunk_finish = hh_min(g_era.finish_date, finish_date)
                site = (
                    sess.query(Site)
                    .join(SiteGEra)
                    .filter(SiteGEra.g_era == g_era, SiteGEra.is_physical == true())
                    .one()
                )

                ds = GDataSource(
                    sess, chunk_start, chunk_finish, forecast_dt, g_era, caches, None
                )

                titles = ["MPRN", "Site Code", "Site Name", "Account", "From", "To", ""]

                output_line = [
                    g_supply.mprn,
                    site.code,
                    site.name,
                    ds.account,
                    hh_format(ds.start_date),
                    hh_format(ds.finish_date),
                    "",
                ]

                contract_titles = g_contract_func(
                    caches, g_era.g_contract, "virtual_bill_titles"
                )()
                titles.extend(contract_titles)

                g_contract_func(caches, g_era.g_contract, "virtual_bill")(ds)
                bill = ds.bill

                for title in contract_titles:
                    if title in bill:
                        output_line.append(csv_make_val(bill[title]))
                        del bill[title]
                    else:
                        output_line.append("")

                for k in sorted(bill.keys()):
                    output_line.extend([k, bill[k]])

                if titles != prev_titles:
                    prev_titles = titles
                    writer.writerow([str(v) for v in titles])
                writer.writerow(output_line)
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    g_supply_id = req_int("g_supply_id")
    file_name = "g_supply_virtual_bill_" + str(g_supply_id) + ".csv"
    start_date = req_date("start")
    finish_date = req_date("finish")
    args = (g_supply_id, file_name, start_date, finish_date, g.user)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import json
from decimal import Decimal
from io import BytesIO
from sqlalchemy import select

from chellow.models import Contract, RateScript
from chellow.rate_server import download
from chellow.utils import ct_datetime, hh_format, to_utc


def ccl(data_source):
    try:
        ccl_cache = data_source.caches["ccl"]
    except KeyError:
        ccl_cache = data_source.caches["ccl"] = {}

    for hh in data_source.hh_data:
        try:
            rate = ccl_cache[hh["start-date"]]
        except KeyError:
            hh_start = hh["start-date"]
            rates = data_source.non_core_rate("ccl", hh_start)
            rate = ccl_cache[hh_start] = float(rates["ccl_gbp_per_msp_kwh"])

        hh["ccl-kwh"] = hh["msp-kwh"]
        hh["ccl-rate"] = rate
        hh["ccl-gbp"] = hh["msp-kwh"] * rate


def rate_server_import(sess, log, set_progress, s, paths):
    log("Starting to check for new electricity CCL rates")
    contract_name = "ccl"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess,
            contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1996, 4, 1)),
            None,
            {},
        )
        sess.commit()
    contract_props = contract.make_properties()

    if not contract_props.get("enabled", False):
        log(
            "The automatic importer is disabled. To enable it, edit "
            "the contract properties to set 'enabled' to true."
        )
        return

    year_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year, utility, rate_type, file_name = path
            if utility == "electricity" and rate_type == "ccl":
                try:
                    fl_entries = year_entries[year]
                except KeyError:
                    fl_entries = year_entries[year] = {}

                fl_entries[file_name] = url

    for year, year_files in sorted(year_entries.items()):
        year_start = to_utc(ct_datetime(year, 4, 1))
        if year_start < contract.start_rate_script.start_date:
            continue
        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == year_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, year_start, {})

        if len(year_files) > 0:
            file_name, url = sorted(year_files.items())[-1]

            rs_script = rs.make_script()
            if rs_script.get("a_file_name") != file_name:
                log(
                    f"Found new file {file_name} for rate script starting "
                    f"{hh_format(year_start)}"
                )
                script = {"a_file_name": file_name}
                server_j = json.load(BytesIO(download(s, url)))
                script["ccl_gbp_per_msp_kwh"] = Decimal(server_j["ccl_gbp_per_msp_kwh"])
                rs.update(script)
                log(f"Updated CCL rate script for {hh_format(year_start)}")

    log("Finished CCL rates")
    sess.commit()
